from dexy.dexy_filter import DexyFilter
import markdown

class MarkdownFilter(DexyFilter):
    INPUT_EXTENSIONS = ['.*']
    OUTPUT_EXTENSIONS = ['.html']
    ALIASES = ['markdown']

    def process_text(self, input_text):
        return markdown.markdown(input_text)





import dexy.filters.blog_filters
import re
import xmlrpclib

class WordPressFilter(dexy.filters.blog_filters.BlogFilter):
    """
    Posts to a WordPress blog.

    Create a file called wp-config.json in the root of your project directory.
    This should include your blog's XMLRPC endpoint URL, your username and your
    password in a JSON object.

    For example:

    {
       "user" : "author",
       "pass" : "password",
       "xmlrpc_url" : "http://blog.dexy.it:80/xmlrpc.php"
    }

    """
    ALIASES = ['wp']
    BLOG_CONFIG_FILE = 'wp-config.json'
    MIME_TYPES = {
        'png' : 'image/png',
        'jpg' : 'image/jpeg',
        'jpeg' : 'image/jpeg',
        'aiff' : 'audio/x-aiff',
        'wav' : 'audio/x-wav',
        'wave' : 'audio/x-wav',
        'mp3' : 'audio/mpeg'
    }
    BLOG_ID = 0

    def initialize_api(self):
        api = xmlrpclib.ServerProxy(self.blog_conf["xmlrpc_url"], verbose=False)
        #print api.system.listMethods()
        return api

    def content_dict(self, api, input_text):
        input_text = self.upload_files_and_replace_links(api, input_text)
        content = { 'title' : self.post_conf['title'], 'description' : input_text}
        return content

    def new_post(self, api, input_text):
        post_id = api.metaWeblog.newPost(
            self.BLOG_ID,
            self.username,
            self.password,
            self.content_dict(api, input_text),
            self.post_conf['publish']
        )
        return post_id

    def update_post(self, api, input_text, post_id):
        api.metaWeblog.editPost(
            post_id,
            self.username,
            self.password,
            self.content_dict(api, input_text),
            self.post_conf['publish']
        )

    def upload_files_and_replace_links(self, api, input_text):
        url_cache = {}

        def upload_files_to_wp(regexp, input_text):
            for t in re.findall(regexp, input_text):
                if url_cache.has_key(t[1]):
                    url = url_cache[t[1]]
                    self.log.info("using cached url %s %s" % (t[1], url))
                else:
                    f = open(t[1], 'rb')
                    image_base_64 = xmlrpclib.Binary(f.read())
                    f.close()

                    upload_file = {
                        'name' : t[1].split("/")[1],
                        'type' : self.MIME_TYPES[t[2]], # *should* raise error if not on whitelist
                        'bits' : image_base_64,
                        'overwrite' : 'true'
                    }
                    upload_result = api.wp.uploadFile(
                        self.BLOG_ID,
                        self.username,
                        self.password,
                        upload_file
                    )
                    url = upload_result['url']
                    url_cache[t[1]] = url
                    self.log.info("uploaded %s to %s" % (t[1], url))

                replace_string = t[0].replace(t[1], url)
                input_text = input_text.replace(t[0], replace_string)
            return input_text

        input_text = upload_files_to_wp('(<img src="(artifacts/.+\.(\w{2,4}))")', input_text)
        input_text = upload_files_to_wp('(<embed src="(artifacts/.+\.(\w{2,4}))")', input_text)
        input_text = upload_files_to_wp('(<audio src="(artifacts/.+\.(\w{2,4}))")', input_text)
        return input_text



import dexy.filters.blog_filter
import posterous # http://github.com/nureineide/posterous-python

class PosterousFilter(dexy.filters.blog_filter.BlogFilter):
    """
    IN DEVELOPMENT. Post to a posterous blog. (Due to posterous' stripping out
    formatting, is difficult to apply syntax highlighting in usual way.)
    """
    ALIASES = ['posterous']
    BLOG_CONFIG_FILE = 'posterous-config.json'

    def initialize_api(self):
        user = self.blog_conf['user']
        password = self.blog_conf['pass']
        return posterous.API(user, password)

    def new_post(self, api, input_text):
        post = api.new_post(title = self.post_conf['title'], body = input_text)
        return post.id

    def update_post(self, api, input_text, post_id):
        api.update_post(
            post_id = post_id,
            title = self.post_conf['title'],
            body = input_text
        )





from dexy.dexy_filter import DexyFilter
import os
import re

class FilenameFilter(DexyFilter):
    """Generate random filenames to track provenance of data."""
    ALIASES = ['fn']

    def process_text(self, input_text):
        # TODO this should not match more than two dashes
        for m in re.finditer("dexy--(\S+)\.([a-z]+)", input_text):
            local_key = m.groups()[0]
            ext = m.groups()[1]

            parent_dir = os.path.dirname(self.artifact.name)
            key = os.path.join(parent_dir, local_key)

            key_with_ext = "%s.%s" % (key, ext)
            key_with_ext_with_dexy = "%s|dexy" % key_with_ext

            if key_with_ext in self.artifact.inputs().keys():
                artifact = self.artifact.inputs()[key_with_ext]
                self.log.debug("[fn] existing key %s in artifact %s links to file %s" %
                          (key_with_ext, self.artifact.key, artifact.filename()))
            elif key_with_ext_with_dexy in self.artifact.inputs().keys():
                artifact = self.artifact.inputs()[key_with_ext_with_dexy]
                self.log.debug("[fn] existing key %s in artifact %s links to existing file %s" %
                          (key_with_ext, self.artifact.key, artifact.filename()))
            else:
                self.log.debug("[fn] could not find match for %s" % (key_with_ext))
                artifact = self.artifact.add_additional_artifact(key_with_ext, ext)
                self.log.debug("[fn] created new artifact %s ; links to new file %s" %
                          (key_with_ext, artifact.filename()))

            input_text = input_text.replace(m.group(), artifact.filename())

        # Hack to replace __ with -- in case we want to document how to use this
        # filter, we can't use -- because this will be acted upon.
        for m in re.finditer("dexy__(.+)\.([a-z]+)", input_text):
            key = m.groups()[0]
            ext = m.groups()[1]
            replacement_key = "dexy--%s.%s" % (key, ext)
            input_text = input_text.replace(m.group(), replacement_key)

        return input_text



from dexy.constants import Constants
import dexy.reporter
import dexy
import dexy.artifact
import dexy.dexy_filter
import inspect
import os
import sys

NULL_LOGGER = Constants.NULL_LOGGER
INSTALL_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

def database_classes(log=NULL_LOGGER):
    """
    Return a dict whose keys are database class names and whose values are the
    corresponding classes.
    """
    database_classes = {}

    d = os.path.join(INSTALL_DIR, 'dexy', 'databases')
    for f in os.listdir(d):
        if f.endswith(".py") and f not in ["base.py", "__init__.py"]:
            log.debug("Loading databases in %s" % os.path.join(d, f))
            basename = f.replace(".py", "")
            module = "dexy.databases.%s" % basename
            try:
                __import__(module)
            except ImportError as e:
                log.warn("database defined in %s are not available: %s" % (module, e))

            if not sys.modules.has_key(module):
                continue

            mod = sys.modules[module]

            for k in dir(mod):
                klass = mod.__dict__[k]
                if inspect.isclass(klass) and not (klass == dexy.database.Database) and issubclass(klass, dexy.database.Database):
                    if database_classes.has_key(k):
                        raise Exception("duplicate database class name %s called from %s in %s" % (k, f, d))
                    database_classes[klass.__name__] = klass

    return database_classes

def artifact_classes(log=NULL_LOGGER):
    """
    Return a dict whose keys are artifact class names and whose values are the
    corresponding classes.
    """
    artifact_classes = {}

    d = os.path.join(INSTALL_DIR, 'dexy', 'artifacts')
    for f in os.listdir(d):
        if f.endswith(".py") and f not in ["base.py", "__init__.py"]:
            log.debug("Loading artifacts in %s" % os.path.join(d, f))
            basename = f.replace(".py", "")
            module = "dexy.artifacts.%s" % basename
            try:
                __import__(module)
            except ImportError as e:
                log.warn("artifact defined in %s are not available: %s" % (module, e))

            if not sys.modules.has_key(module):
                continue

            mod = sys.modules[module]

            for k in dir(mod):
                klass = mod.__dict__[k]
                if inspect.isclass(klass) and not (klass == dexy.artifact.Artifact) and issubclass(klass, dexy.artifact.Artifact):
                    if not artifact_classes.has_key(k):
                        artifact_classes[klass.__name__] = klass

    return artifact_classes

def get_filter_by_name(name, filter_list=None):
    if not filter_list:
        # populate the filter list ourselves if not supplied...
        filter_list = filters()

    classes = [k for k in filter_list if k.__name__ == name]
    if len(classes) == 0:
        raise Exception("no filter class %s found" % name)
    return classes[0]

def get_filter_for_alias(alias, filter_list=None):
    if not filter_list:
        # populate the filter list ourselves if not supplied...
        filter_list = filters()

    if filter_list.has_key(alias):
        return filter_list[alias]
    elif alias.startswith("-") or alias.startswith("alias-") or alias.startswith("al-") or alias in ['al', 'alias']:
        return dexy.dexy_filter.DexyFilter
    else:
        raise Exception("filter alias '%s' not found or not available" % alias)

def filters(log=NULL_LOGGER):
    """
    Returns a dict whose keys are all supported filter alises and whose values
    are the corresponding filter classes.
    """
    dexy_filters = ('dexy.filters', os.path.join(dexy.__path__[0], 'filters'))
    proj_filters = ('filters', os.path.abspath(os.path.join(os.curdir, 'filters')))
    user_filters = ('dexy_filters', os.path.expanduser(os.path.join('~', 'dexy_filters')))

    filter_dirs = []

    if os.path.exists(proj_filters[1]):
        init_py_file = os.path.join(proj_filters[1], "__init__.py")
        path = os.path.abspath(os.curdir)
        if not os.path.exists(init_py_file):
            print "You need to create a __init__.py file in", proj_filters[1], "in order for filters to be available"
        elif not path in sys.path:
            print "Adding", path, "to python sys.path so your custom filters in", proj_filters[1], "will be available"
            sys.path.append(path)

    if os.path.exists(user_filters[1]):
        init_py_file = os.path.join(user_filters[1], "__init__.py")
        path = os.path.expanduser('~')
        if not os.path.exists(init_py_file):
            print "You need to create a __init__.py file in", user_filters[1], "in order for filters to be available"
        elif not path in sys.path:
            print "Adding", path, "to python sys.path so your custom filters in", user_filters[1], "will be available"
            sys.path.append(path)

    for pkg, d in [dexy_filters, proj_filters, user_filters]:
         if os.path.exists(d) and (pkg, d) not in filter_dirs:
             filter_dirs.append((pkg, d))

    filters = {}

    for a in dexy.dexy_filter.DexyFilter.ALIASES:
        filters[a] = dexy.dexy_filter.DexyFilter

    for pkg, d in filter_dirs:
        log.info("Automatically loading all %s found in %s" % (pkg, d))
        for f in os.listdir(d):
            if f.endswith(".py") and f not in ["base.py", "__init__.py"]:
                log.info("Loading filters in %s" % os.path.join(d, f))
                basename = f.replace(".py", "")
                modname = "%s.%s" % (pkg, basename)

                try:
                    __import__(modname)
                except ImportError as e:
                    log.warn("filters defined in %s are not available: %s" % (modname, e))

                if not sys.modules.has_key(modname):
                    continue

                mod = sys.modules[modname]

                for k in dir(mod):
                    klass = mod.__dict__[k]

                    is_class = inspect.isclass(klass)

                    if is_class and issubclass(klass, dexy.dexy_filter.DexyFilter) and (klass.__module__ == modname):
                        if not klass.ALIASES:
                            log.info("class %s is not available because it has no aliases" % klass.__name__)
                        elif not klass.executable_present():
                            log.info("class %s is not available because %s not found" %
                                          (klass.__name__, klass.executable()))
                        elif not klass.enabled():
                            log.info("class %s is not available because it is not enabled" %
                                          (klass.__name__))
                        else:
                            for a in klass.ALIASES:
                                if filters.has_key(a):
                                    raise Exception("duplicate key %s called from %s in %s" % (a, k, f))
                                filters[a] = klass
                                log.info("registered alias %s for class %s" % (a, k))
        log.info("...finished loading filters from %s" % d)
    return filters

def reporters(log=NULL_LOGGER):
    """
    Returns a dict of reporter names and classes.
    """
    # Reporters that come with dexy are installed in the reporters/ subdir:
    d1 = os.path.abspath(os.path.join(INSTALL_DIR, 'reporters'))

    # Custom reporters for a project may be placed in a reporters/ dir in the project:
    d2 = os.path.abspath(os.path.join(os.curdir, 'reporters'))

    if d1 == d2 or not os.path.exists(d2):
        reporter_dirs = [d1]
    else:
        reporter_dirs = [d1,d2]

    reporters = {}
    for d in reporter_dirs:
        log.info("Loading reporters in dir %s" % d)
        for f in os.listdir(d):
            if f.endswith(".py") and f not in ["base.py", "__init__.py"]:
                log.info("Loading reporters in %s" % os.path.join(d, f))
                basename = f.replace(".py", "")
                module = "reporters.%s" % basename

                try:
                    __import__(module)
                except ImportError as e:
                    log.warn("reporters defined in %s are not available: %s" % (module, e))

                if not sys.modules.has_key(module):
                    continue

                mod = sys.modules[module]

                for k in dir(mod):
                    klass = mod.__dict__[k]
                    if inspect.isclass(klass) and not (klass == dexy.reporter.Reporter) and issubclass(klass, dexy.reporter.Reporter):
                        reporters[klass.__name__] = klass
    return reporters

def reports_dirs(log=NULL_LOGGER):
    """
    Returns a list of all directories which reporters declare that they use.
    """
    return [r.REPORTS_DIR for r in reporters(log).values() if r.REPORTS_DIR and not r.REPORTS_DIR.startswith("logs/")]



from dexy.dexy_filter import DexyFilter
import gdata.photos.service
import json
import mimetypes
import os

class PicasaFilter(DexyFilter):
    # TODO can support other input formats, but they get converted to jpeg, need to test implications
    # However within Dexy you should be able to control what format to produce.
    INPUT_EXTENSIONS = ['.jpg']
    OUTPUT_EXTENSIONS = ['.json']
    ALIASES = ['picasa']
    CONFIG_FILE = 'picasa-config.json'
    DEFAULT_CONF = {
        "access" : 'public', # You can set this to access : private in your config.
        "license" : None # Need to figure out how to specify license info.
    }

    def load_config(self):
        if not os.path.exists(self.CONFIG_FILE):
            raise Exception("Could not find config file called %s" % self.CONFIG_FILE)

        self.conf = self.DEFAULT_CONF
        f = open(self.CONFIG_FILE, "r")
        self.conf.update(json.load(f))
        f.close()

    def process_text(self, input_text):
        self.load_config()

        client = gdata.photos.service.PhotosService()

        client.email = self.conf['email']
        client.password = self.conf['password']
        client.source = 'dexy'
        client.ProgrammaticLogin()

        album_name = "Dexy: %s" % self.conf['project']

        # If an album for this project already exists, get it. Otherwise, create a new one.
        albums = client.GetUserFeed('album')

        username = albums.user.text
        matching_albums = [album for album in albums.entry if album.title.text == album_name]
        if len(matching_albums) == 0:
            print "Creating a new %s Picasa album called %s" % (self.conf['access'], album_name)
            album = client.InsertAlbum(
                    album_name,
                    "Files automatically uploaded by Dexy for the %s project." % self.conf['project'],
                    access=self.conf['access']
                    )
        else:
            album = matching_albums[0]


        photo_key = self.artifact.key.replace("/", "--")
        filename = self.artifact.previous_artifact_filepath
        mimetype = mimetypes.guess_type(filename)[0]

        # If a photo with this title in this album already exists, replace it. Otherwise, start from scratch.
        photos = client.GetUserFeed("photo")
        photo = None
        for photo in photos.entry:
            photo_in_album = album.id.text.split("/")[-1] == photo.albumid.text
            photo_tagged_dexy = photo.media.keywords and 'dexydocs' in photo.media.keywords.text
            title_matches = photo.title.text == photo_key

            if photo_in_album and photo_tagged_dexy and title_matches:
                break
            else:
                photo = None

        if photo:
            client.UpdatePhotoBlob(photo, filename, content_type=mimetype)
        else:
            photo = client.InsertPhotoSimple(album, photo_key, "", filename, content_type=mimetype)
            photo.media.keywords = gdata.media.Keywords()
            photo.media.keywords.text = "dexydocs"
            # TODO support geolocation to show where research is taking place
            client.UpdatePhotoMetadata(photo)

        photo_data = {
            "id" : photo.gphoto_id.text,
            "height" : photo.height.text,
            "width" : photo.width.text,
            "src" : photo.content.src,
            "url" : photo.GetHtmlLink().href
        }

        return json.dumps(photo_data)




import dexy.introspect

INIT_ARGS = {
        "SqliteDatabase" : { "logsdir" : None, "dbfile" : None }
}

REQUIRED_METHODS = [
    'persist',
    'max_batch_id',
    'next_batch_id',
    'next_batch_order',
    'append_artifacts',
    'update_artifact',
    'references_for_batch_id'
]

def test_db_api():
    database_classes = dexy.introspect.database_classes()
    for klass_name, klass in database_classes.iteritems():
        print "testing", klass_name
        args = INIT_ARGS[klass_name]
        print "initializing db with", args
        db = klass(**args)
        for req in REQUIRED_METHODS:
            assert hasattr(db, req), "db class %s must implement a %s method" % (klass_name, req)




from dexy.dexy_filter import DexyFilter
from dexy.dexy_filter import DexyFilterException
from dexy.filters.templating_plugins import *
from jinja2.exceptions import TemplateSyntaxError
from jinja2.exceptions import UndefinedError
import jinja2
import sys
import re
import traceback

class TemplateFilterException(DexyFilterException):
    pass

class JinjaFilterException(TemplateFilterException):
    pass

class TemplateFilter(DexyFilter):
    """
    Base class for templating system filters such as JinjaFilter. Templating
    systems are used to make generated artifacts available within documents.

    Plugins are used to prepare content.
    """
    ALIASES = ['template']
    FINAL = True
    PLUGINS = [
        ClippyHelperTemplatePlugin,
        DexyVersionTemplatePlugin,
        GlobalsTemplatePlugin,
        InputsTemplatePlugin,
        PrettyPrinterTemplatePlugin,
        PygmentsStylesheetTemplatePlugin,
        PythonBuiltinsTemplatePlugin,
        RegularExpressionsTemplatePlugin,
        SubdirectoriesTemplatePlugin,
        VariablesTemplatePlugin
        ]

    def run_plugins(self):
        env = {}
        for plugin_class in self.plugins():
            plugin = plugin_class(self)
            new_env_vars = plugin.run()
            if any(v in env.keys() for v in new_env_vars):
                raise Exception("trying to add new keys %s, already have %s" % (u", ".join(new_env_vars.keys()), u", ".join(env.keys())))
            env.update(new_env_vars)
        return env

    def plugins(self):
        if self.artifact.args.has_key('plugins'):
            plugin_names = self.artifact.args['plugins']
            dict_items = sys.modules[self.__module__].__dict__
            return [dict_items[name] for name in plugin_names]
        else:
            return self.PLUGINS

    def process_text(self, input_text):
        """
        Overwrite this in your subclass, or, perhaps better, overwrite
        process() and write output directly to artifact file.
        """
        template_data = self.run_plugins()
        return input_text % template_data

class JinjaTextFilter(TemplateFilter):
    """
    Runs the Jinja templating engine. Uses process_text and returns text rather
    than writing to file.
    """
    ALIASES = ['jinjatext']

    def process_text(self, input_text):
        env = self.setup_jinja_env()
        template_data = self.run_plugins()
        try:
            template = env.from_string(input_text)
            return template.render(template_data)
        except (TemplateSyntaxError, UndefinedError, TypeError) as e:
            self.handle_jinja_exception(e)

    def setup_jinja_env(self):
        if self.artifact.ext == ".tex":
            self.log.debug("changing jinja tags to << >> etc. for %s" % self.artifact.key)
            env = jinja2.Environment(
                block_start_string = '<%',
                block_end_string = '%>',
                variable_start_string = '<<',
                variable_end_string = '>>',
                comment_start_string = '<#',
                comment_end_string = '#>',
                undefined = jinja2.StrictUndefined
                )
        else:
            env = jinja2.Environment(
                undefined = jinja2.StrictUndefined
            )
        return env

    def handle_jinja_exception(self, e):
        result = []
        br = "=================================================="
        s = "    "

        if isinstance(e, UndefinedError) or isinstance(e, TypeError):
            # try to get the line number
            m = re.search(r"File \"<template>\", line ([0-9]+), in top\-level template code", traceback.format_exc())
            if m:
                e.lineno = int(m.groups()[0])
            else:
                raise Exception("Unable to parse line number from %s" % traceback.format_exc())

        result.append("""There is a problem with %(key)s
        \nA problem was detected at line %(lineno)s of %(workfile)s
        """ % {'key' : self.artifact.key, 'lineno' : e.lineno, 'workfile' : self.artifact.previous_artifact_filepath })

        if isinstance(e, UndefinedError):
            result.append("WARNING: line number may not be accurate, check elsewhere in your file if the excerpt does not contain the undefined item from the stack trace.")

        input_lines = self.artifact.input_text().splitlines()

        result.append(br)

        # print context before line with problem, if available
        if e.lineno >= 3:
            result.append(s + input_lines[e.lineno-3])
        if e.lineno >= 2:
            result.append(s + input_lines[e.lineno-2])

        # this is the line that has the problem
        result.append(">>> %s" % input_lines[e.lineno-1])

        # print context after line with problem, if available
        if len(input_lines) > e.lineno:
            result.append(s + input_lines[e.lineno-0])
        if len(input_lines) > (e.lineno + 1):
            result.append(s + input_lines[e.lineno+1])

        result.append(br)
        result.append("The error is: %s" % e.message)

        result.append(traceback.format_exc())

        # Exception constructors don't like unicode, so print error messsage to
        # STDOUT then raise an exception.
        print u"\n".join(result)
        raise JinjaFilterException("An error has occurred while processing %s" % self.artifact.key)

class JinjaFilter(JinjaTextFilter):
    """
    Runs the Jinja templating engine on your document to incorporate dynamic
    content.
    """
    ALIASES = ['jinja']
    TAGS = ['template']

    def process(self):
        env = self.setup_jinja_env()
        template_data = self.run_plugins()
        try:
            template = env.from_string(self.artifact.input_text())
            template.stream(template_data).dump(self.artifact.filepath(), encoding="utf-8")
        except (TemplateSyntaxError, UndefinedError, TypeError) as e:
            self.handle_jinja_exception(e)



class Database(object):
    """
    The Database class stores metadata around dexy batches/runs. This metadata
    is used for reporting and can be used as a version history of the documents
    you process with dexy.
    The database can also be used to parallelize/distribute the processing of
    dexy runs and to store ctime/mtime/inode data to speed up cache detection.
    """

    def __init__(self, **kwargs):
        raise Exception("not implemented")

    def append_artifact(self, artifact):
        self.append_artifacts([artifact])



from StringIO import StringIO
import sys
import os
import shutil
import tempfile

class tempdir():
    def __enter__(self):
        self.tempdir = tempfile.mkdtemp()
        self.location = os.path.abspath(os.curdir)
        os.chdir(self.tempdir)

    def __exit__(self, type, value, traceback):
        os.chdir(self.location)
        shutil.rmtree(self.tempdir)

class divert_stdout():
    def __enter__(self):
        self.old_stdout = sys.stdout
        self.my_stdout = StringIO()
        sys.stdout = self.my_stdout
        return self.my_stdout

    def __exit__(self, type, value, traceback):
        sys.stdout = self.old_stdout
        self.my_stdout.close()




from dexy.constants import Constants
from ordereddict import OrderedDict
import sqlite3
import dexy.database
import json
import os

class SqliteDatabase(dexy.database.Database):
    """
    The Database class stores metadata around dexy batches/runs. This metadata
    is used for reporting and can be used as a version history of the documents
    you process with dexy.
    The database can also be used to parallelize/distribute the processing of
    dexy runs and to store ctime/mtime/inode data to speed up cache detection.
    """

    FIELDS = [
            ("batch_id" , "int"),
            ("batch_order" , "int"),
            ("elapsed" , "int"),
            ("hashstring" , "text"),
            ("source" , "text")
            ]

    def create_table_sql(self):
        sql = "create table artifacts (id text primary key, %s)"
        fields = ["%s %s" % k for k in self.FIELDS]
        fields.extend("%s text" % k for k in self.whitelist_keys)
        return sql % (", ".join(fields))

    def create_table(self):
        try:
            self.conn.execute(self.create_table_sql())
        except sqlite3.OperationalError:
            pass

    def __init__(self, logsdir=Constants.DEFAULT_LDIR, dbfile=Constants.DEFAULT_DBFILE):
        self.logsdir = logsdir or ""
        if dbfile:
            filename = os.path.join(self.logsdir, dbfile)
            self.filename = filename
            # TODO test that schema hasn't changed
        else:
            # use in-memory db
            self.filename = ":memory:"

        self.conn = sqlite3.connect(self.filename)
        self.conn.row_factory = sqlite3.Row

        self.field_keys = [k[0] for k in self.FIELDS]
        self.whitelist_keys = sorted(Constants.ARTIFACT_HASH_WHITELIST)
        self.field_names = ['id'] + self.field_keys + self.whitelist_keys
        self.create_table()
        self.batch_orders = {}

    def persist(self):
        self.conn.commit()

    def max_batch_id(self):
        return self.conn.execute("select max(batch_id) from artifacts").fetchone()[0] or 0

    def next_batch_id(self):
        return self.max_batch_id() + 1

    def max_batch_order(self, batch_id, incr=False):
        # using a hash for calculating batch order since hitting the DB each
        # time is too slow. should be getting this from topo sort? should be
        # per-artifact rather than per-document? This isn't used for anything
        # currently, so not a big deal.
        if not self.batch_orders.has_key(batch_id):
            self.batch_orders[batch_id] = 0
        elif incr:
            self.batch_orders[batch_id] += 1
        return self.batch_orders[batch_id]

    def next_batch_order(self, batch_id):
        return self.max_batch_order(batch_id, True)

    def get_attributes_for_artifact(self, artifact):
        hd = artifact.hash_dict()
        values = [artifact.unique_key()]

        # add attrs not in hash whitelist
        values.extend(getattr(artifact, k) for k in self.field_keys)

        # get attrs from the hash whitelist, converting any OrderedDicts into JSON
        artifact_values = (hd.get(k, None) for k in self.whitelist_keys)
        convert_to_json = lambda v: isinstance(v, OrderedDict) and json.dumps(v) or v
        values.extend(convert_to_json(v) for v in artifact_values)

        return values

    def append_artifacts(self, artifacts):
        qs = ("?," * len(self.field_names))[:-1]
        sql = "INSERT INTO artifacts VALUES (%s)" % qs
        for a in artifacts:
            self.conn.execute(sql, self.get_attributes_for_artifact(a))

    def update_artifact(self, artifact):
        values = self.get_attributes_for_artifact(artifact)[1:]
        values.append(artifact.unique_key())
        sql = "UPDATE artifacts SET %s where id = ?" % (", ".join("%s = ? " % k for k in self.field_names[1:]))
        self.conn.execute(sql, values)

    def artifact_row(self, artifact):
        """Returns the db row corresponding to an artifact."""
        sql = "SELECT * from artifacts where id = ?"
        return self.conn.execute(sql, (artifact.unique_key(),)).fetchone()

    def references_for_batch_id(self, batch_id=None):
        """
        Return information for a given batch.
        """
        if not batch_id:
            # use most recent batch
            batch_id = self.max_batch_id()
        sql = "SELECT * from artifacts where batch_id = ?"
        return self.conn.execute(sql, (batch_id,)).fetchall()

    def all(self):
        return self.conn.execute("SELECT * from artifacts").fetchall()



from dexy.artifact import Artifact
from dexy.filters.templating_filters import TemplateFilter
from dexy.filters.templating_filters import TemplatePlugin

class TestTemplatePlugin(TemplatePlugin):
    def run(self):
        return {'a' : 1}

class TestTemplateFilter(TemplateFilter):
    PLUGINS = [TestTemplatePlugin]

def test_template_filter():
    test_filter = TestTemplateFilter()
    test_filter.artifact = Artifact()
    env = test_filter.run_plugins()
    assert env.has_key('a')
    assert env['a'] == 1

    s = test_filter.process_text("The value of a is %(a)s")
    assert s == "The value of a is 1"

def test_template_filter_with_custom_plugins():
    test_filter = TestTemplateFilter()
    test_filter.artifact = Artifact()
    test_filter.artifact.args['plugins'] = ['TestTemplatePlugin']
    env = test_filter.run_plugins()
    assert env.has_key('a')
    assert env['a'] == 1

    s = test_filter.process_text("The value of a is %(a)s")
    assert s == "The value of a is 1"



from PythonMagick import Image
from PythonMagick import _PythonMagick
from dexy.dexy_filter import DexyFilter

class PythonMagickFilter(DexyFilter):
    """Base class which reads previous artifact, calls a do_magick method,
    saves the output. Can be inherited. Default action is to flip an image,
    pointless in itself but a useful demo to know things are working."""
    OUTPUT_EXTENSIONS = [".png"]
    INPUT_EXTENSIONS = [".png"]
    ALIASES = ['pm-flip']
    BINARY = True

    def do_magick(self, image):
        image.flip()

    def process(self):
        wf = str(self.artifact.previous_artifact_filepath) # complains when unicode
        image = Image()
        image.read(wf)
        self.do_magick(image)
        image.write(self.artifact.filepath())

class MagickFlopFilter(PythonMagickFilter):
    ALIASES = ['pm-flop']

    def do_magick(self, image):
        image.flop()

class MagickGridFilter(PythonMagickFilter):
    ALIASES = ['pm-grid']
    # This does not work due to PythonMagick issues, maybe fixed in later version of PM?
    def do_magick(self, image):
        for i in xrange(10, 100, 10):
            for j in xrange(10, 100, 10):
                print "i: %s j: %s" % (i, j)
                dl = _PythonMagick.DrawableLine(0, 0, i, j)
                image.draw(dl)





from ordereddict import OrderedDict
from pygments.styles import get_all_styles
import json
import os
import pprint
import pygments
import pygments.formatters
import re
import urllib
from dexy.version import Version

class TemplatePlugin(object):
    def __init__(self, filter_instance):
        self.filter_instance = filter_instance

    def run(self):
        return {}

class DexyVersionTemplatePlugin(TemplatePlugin):
    def run(self):
        return { "dexy_version" : Version.VERSION }

class PrettyPrinterTemplatePlugin(TemplatePlugin):
    def run(self):
        return { 'pformat' : pprint.pformat}

class RegularExpressionsTemplatePlugin(TemplatePlugin):
    def run(self):
        return { 're_match' : re.match, 're_search' : re.search}

class PythonBuiltinsTemplatePlugin(TemplatePlugin):
    # Intended to be all builtins that make sense to run within a document.
    PYTHON_BUILTINS = [abs, all, any, bin, bool, bytearray, callable, chr,
        cmp, complex, dict, dir, divmod, enumerate, filter, float, format,
        hex, id, int, isinstance, issubclass, iter, len, list, long, map, hasattr,
        max, min, oct, ord, pow, range, reduce, repr, reversed, round,
        set, slice, sorted, str, sum, tuple, xrange, zip]

    def run(self):
        return dict((f.__name__, f) for f in self.PYTHON_BUILTINS)

class PygmentsStylesheetTemplatePlugin(TemplatePlugin):
    def highlight(self, text, lexer_name, fmt = 'html', noclasses = False):
        formatter_options = { "noclasses" : noclasses }
        lexer = pygments.lexers.get_lexer_by_name(lexer_name)
        formatter = pygments.formatters.get_formatter_by_name(fmt, **formatter_options)
        return pygments.highlight(text, lexer, formatter)

    def run(self):
        pygments_stylesheets = {}
        for style_name in get_all_styles():
            for formatter_class in [pygments.formatters.LatexFormatter, pygments.formatters.HtmlFormatter]:
                pygments_formatter = formatter_class(style=style_name)
                style_info = pygments_formatter.get_style_defs()

                for fn in pygments_formatter.filenames:
                    ext = fn.split(".")[1]
                    if ext == 'htm':
                        ext = 'css' # swap the more intuitive '.css' for the unlikely '.htm'
                    key = "%s.%s" % (style_name, ext)
                    pygments_stylesheets[key] = style_info
        return {'pygments' : pygments_stylesheets, 'highlight' : self.highlight }

class SubdirectoriesTemplatePlugin(TemplatePlugin):
    def run(self):
        # The directory containing the document to be processed.
        doc_dir = os.path.dirname(self.filter_instance.artifact.name)

        # Get a list of subdirectories under this document's directory.
        subdirectories = [d for d in sorted(os.listdir(os.path.join(os.curdir, doc_dir))) if os.path.isdir(os.path.join(os.curdir, doc_dir, d))]
        return {'subdirectories' : subdirectories}

class VariablesTemplatePlugin(TemplatePlugin):
    def run(self):
        if self.filter_instance.artifact.args.has_key('variables'):
            return self.filter_instance.artifact.args['variables']
        else:
            return {}

class GlobalsTemplatePlugin(TemplatePlugin):
    def run(self):
        if self.filter_instance.artifact.controller_args.has_key('globals'):
            return self.filter_instance.artifact.controller_args['globals']
        else:
            return {}

class InputsTemplatePlugin(TemplatePlugin):
    def load_sort_json_data(self, a):
        try:
            unsorted_json = json.loads(a.output_text())
        except ValueError as e:
            print "unable to load JSON for", a.key
            raise e

        def sort_dict(d):
            od = OrderedDict()
            for k in sorted(d.keys()):
                v = d[k]
                if isinstance(v, dict) or isinstance(v, OrderedDict):
                    od[k] = sort_dict(v)
                else:
                    od[k] = v
            return od

        return sort_dict(unsorted_json)

    def run(self):
        d_hash = {}
        a_hash = {}

        name = self.filter_instance.artifact.name
        inputs = self.filter_instance.artifact.inputs()

        for key, a in inputs.iteritems():
            keys = a.relative_refs(name)

            # Do any special handling of data
            if a.ext == '.json':
                data = self.load_sort_json_data(a)
            else:
                data = a.data_dict

            for k in keys:
                # Avoid adding duplicate keys
                if a_hash.has_key(k):
                    next

                a_hash[k] = a

                if data.keys() == ['1']:
                    d_hash[k] = data['1']
                else:
                    d_hash[k] = data

        return {
            'a' : a_hash,
            's' : self.filter_instance.artifact,
            'd' : d_hash,
            'f' : self.filter_instance,
        }

class ClippyHelperTemplatePlugin(TemplatePlugin):
    PRE_AND_CLIPPY_STRING = """
<pre>
%s
</pre>%s
    """
    CLIPPY_HELPER_STRING = """
<object classid="clsid:d27cdb6e-ae6d-11cf-96b8-444553540000"
                width="110"
                height="14"
                id="clippy" >
<param name="movie" value="/clippy.swf"/>
<param name="allowScriptAccess" value="always" />
<param name="quality" value="high" />
<param name="scale" value="noscale" />
<param NAME="FlashVars" value="text=%s">
<param name="bgcolor" value="#ffffff">
<embed src="/clippy.swf"
       width="110"
       height="14"
       name="clippy"
       quality="high"
       allowScriptAccess="always"
       type="application/x-shockwave-flash"
       pluginspage="http://www.macromedia.com/go/getflashplayer"
       FlashVars="text=%s"
       bgcolor="#ffffff"
/>
</object>"""

    def run(self):
        return {
            'pre_and_clippy' : self.pre_and_clippy,
            'clippy_helper' : self.clippy_helper
        }

    def pre_and_clippy(self, text):
        return self.PRE_AND_CLIPPY_STRING % (text, self.clippy_helper(text))

    def clippy_helper(self, text):
        if not text or len(text) == 0:
            raise Exception("You passed blank text to clippy helper!")
        quoted_text = urllib.quote(text)
        return self.CLIPPY_HELPER_STRING % (quoted_text, quoted_text)



from dexy.dexy_filter import DexyFilter
import os
import tarfile
import zipfile

class UnprocessedDirectoryArchiveFilter(DexyFilter):
    """
    Create a tgz archive containing the original (unprocessed) files in a directory.
    """
    OUTPUT_EXTENSIONS = [".tgz"]
    ALIASES = ['tgzdir']
    BINARY = True
    FINAL = True

    def process(self):
        dirname = os.path.dirname(self.artifact.name)
        tgz_dir = os.path.join(dirname, self.artifact.args['dir'])

        af = self.artifact.filepath()
        tar = tarfile.open(af, mode="w:gz")

        for fn in os.listdir(tgz_dir):
            fp = os.path.join(tgz_dir, fn)
            self.artifact.log.debug("Adding file %s to archive %s." % (fp, af))
            tar.add(fp, arcname=os.path.join(self.artifact.args['dir'], fn))

        tar.close()

class ArchiveFilter(DexyFilter):
    """
    Creates a .tgz archive of all input documents.

    The use-short-names option will store documents under their short
    (canonical) filenames.
    """
    OUTPUT_EXTENSIONS = [".tgz"]
    ALIASES = ['archive', 'tgz']
    BINARY = True
    FINAL = True

    def process(self):
        if self.artifact.args.has_key('use-short-names'):
            use_short_names = self.artifact.args['use-short-names']
        else:
            use_short_names = False

        af = self.artifact.filepath()
        tar = tarfile.open(af, mode="w:gz")

        # Place files in the archive within a directory with the same name as the archive
        dirname = os.path.splitext(os.path.basename(self.artifact.name))[0]

        for k, a in self.artifact.inputs().items():
            fn = a.filepath()
            if not os.path.exists(fn):
                raise Exception("File %s does not exist!" % fn)
            if use_short_names:
                arcname = a.canonical_filename()
            else:
                arcname = a.long_canonical_filename()

            arcname = os.path.join(self.artifact.relative_path_to_input(a), os.path.basename(arcname))
            arcname = os.path.join(dirname, arcname)

            self.artifact.log.debug("Adding file %s to archive %s as %s" % (fn, af, arcname))
            tar.add(fn, arcname=arcname)
        tar.close()

class ZipArchiveFilter(DexyFilter):
    """
    Creates a .zip archive of all input documents.

    The use-short-names option will store documents under their short
    (canonical) filenames.
    """
    OUTPUT_EXTENSIONS = [".zip"]
    ALIASES = ['zip']
    BINARY = True
    FINAL = True

    def process(self):
        if self.artifact.args.has_key('use-short-names'):
            use_short_names = self.artifact.args['use-short-names']
        else:
            use_short_names = False
        af = self.artifact.filepath()
        zf = zipfile.ZipFile(af, mode="w")
        for k, a in self.artifact.inputs().items():
            fn = a.filepath()
            if not os.path.exists(fn):
                raise Exception("File %s does not exist!" % fn)
            if use_short_names:
                arcname = a.canonical_filename()
            else:
                arcname = a.long_canonical_filename()
            self.artifact.log.debug("Adding file %s to archive %s." % (fn, af))
            zf.write(fn, arcname=arcname)
        zf.close()




from dexy.commands import *
from dexy.constants import Constants
from dexy.tests.utils import divert_stdout
from dexy.tests.utils import tempdir
import dexy.artifact
import dexy.utils
import os

SIMPLE_PY_CONFIG = """
{
   "@simple.py|py" : {
       "contents" : "x=6\\ny=7\\nprint x*y"
    }
}"""

def test_setup_message_without_config():
    with tempdir():
        with divert_stdout() as stdout:
            setup_command()
            assert "You are almost ready" in stdout.getvalue()

def test_setup_message_with_config():
    with tempdir():
        with divert_stdout() as stdout:
            with open(Constants.DEFAULT_CONFIG, "w") as f:
                f.write(SIMPLE_PY_CONFIG)
            setup_command()
            assert "You are now ready" in stdout.getvalue()

def test_commands_reset():
    with tempdir():
        setup_command()
        reset_command()
        assert check_setup(logsdir=Constants.DEFAULT_LDIR, artifactsdir=Constants.DEFAULT_ADIR)

def test_commands_reset_custom_location():
    with tempdir():
        logsdir = "logsx"
        artifactsdir = "artifactsx"
        setup_command(logsdir=logsdir, artifactsdir=artifactsdir)
        assert check_setup(logsdir=logsdir, artifactsdir=artifactsdir)
        assert os.path.exists(artifactsdir)
        assert os.path.exists(logsdir)
        reset_command(logsdir=logsdir, artifactsdir=artifactsdir)
        assert check_setup(logsdir=logsdir, artifactsdir=artifactsdir)

def test_commands_setup_cleanup():
    with tempdir():
        with divert_stdout() as stdout:
            setup_command()
            assert "You are almost ready" in stdout.getvalue()
            assert os.path.exists(Constants.DEFAULT_ADIR)
            assert os.path.exists(Constants.DEFAULT_LDIR)
            assert check_setup(logsdir=Constants.DEFAULT_LDIR, artifactsdir=Constants.DEFAULT_ADIR)

        cleanup_command()
        assert not os.path.exists(Constants.DEFAULT_ADIR)
        assert not os.path.exists(Constants.DEFAULT_LDIR)
        assert not check_setup(logsdir=Constants.DEFAULT_LDIR, artifactsdir=Constants.DEFAULT_LDIR)

def test_commands_dexy_without_setup():
    with tempdir():
        with divert_stdout() as stdout:
            try:
                dexy_command()
                assertFalse # should not get here
            except SystemExit as e:
                assert "Please run 'dexy setup'" in stdout.getvalue()
                assert not check_setup(logsdir=Constants.DEFAULT_LDIR, artifactsdir=Constants.DEFAULT_LDIR)
                assert e.code == 1

def test_commands_reporters():
    with divert_stdout() as stdout:
        reporters_command()
        assert "Output" in stdout.getvalue()

def test_commands_history():
    with tempdir():
        with divert_stdout() as stdout:
            with open(".dexy", "w") as f:
                f.write(SIMPLE_PY_CONFIG)

            setup_command()
            dexy_command()
            dexy_command()

            history_command(filename="simple.py")
            text = stdout.getvalue()
            assert "Dexy found these versions of simple.py" in text
            assert "logs/batch-source-00001/simple.py" in text
            assert "logs/batch-source-00002/simple.py" in text

def test_help_text():
    assert "Available commands for dexy are" in help_text()
    assert "e.g. 'dexy setup --artifactsdir artifacts'" in help_text("setup")
    assert "e.g. 'dexy --artifactsdir artifacts'" in help_text("dexy")

def test_help_command():
    with divert_stdout() as stdout:
        help_command()
        assert stdout.getvalue() == help_text()

def test_help_command_with_on():
    with divert_stdout() as stdout:
        help_command("setup")
        assert stdout.getvalue() == help_text("setup")

def test_dexy_command():
    with tempdir():
        setup_command()
        dexy_command()

def test_dexy_command_with_data():
    with tempdir():
        with open(".dexy", "w") as f:
            f.write(SIMPLE_PY_CONFIG)

        setup_command()
        dexy_command()

        assert os.path.exists("logs/batch-00001.json")
        db = dexy.utils.get_db()
        assert os.path.exists(db.filename)
        for row in db.references_for_batch_id():
            h = row['hashstring']
            assert os.path.exists("artifacts/%s-output.json" % h)
            assert os.path.exists("artifacts/%s-meta.json" % h)

        # Test database.
        reporter = dexy.reporter.Reporter()
        reporter.load_batch_artifacts()
        for a in reporter.artifacts.values():
            assert isinstance(a, dexy.artifact.Artifact)

def test_caching():
    with tempdir():
        with open(".dexy", "w") as f:
            f.write(SIMPLE_PY_CONFIG)
        setup_command()
        dexy_command()

        db = dexy.utils.get_db()
        refs = db.references_for_batch_id()
        assert refs[1]['source'] == 'run'

        dexy_command()

        db = dexy.utils.get_db()
        refs = db.references_for_batch_id()
        assert refs[1]['source'] == 'cache'

        dexy_command(nocache=True)

        db = dexy.utils.get_db()
        refs = db.references_for_batch_id()
        assert refs[1]['source'] == 'run'

def test_commands_filters():
    with divert_stdout() as stdout:
        filters_command()
        assert "looking up filter information..." in stdout.getvalue()
        assert "ZipArchiveFilter (zip)" in stdout.getvalue()
        assert "RubySubprocessStdoutFilter (rb)" in stdout.getvalue()
        assert "SplitLatexFilter (splitlatex) Splits a latex doc into multiple latex docs." in stdout.getvalue()
        assert not "http://dexy.it/docs/filters/tgzdir" in stdout.getvalue()

def test_commands_filters_individual_filter():
    with divert_stdout() as stdout:
        filters_command(alias="tgzdir")
        assert "UnprocessedDirectoryArchiveFilter" in stdout.getvalue()
        assert "Aliases: tgzdir" in stdout.getvalue()
        assert "Create a tgz" in stdout.getvalue()
        assert "http://dexy.it/docs/filters/tgzdir" in stdout.getvalue()

def test_commands_filters_source():
    with divert_stdout() as stdout:
        filters_command(alias="tgzdir", source=True)
        with_color = stdout.getvalue()

    with divert_stdout() as stdout:
        filters_command(alias="tgzdir", source=True, nocolor=True)
        without_color = stdout.getvalue()

    assert len(with_color) > len(without_color)
    # TODO find a way to strip out ansi color codes and assert that text is equal?



from dexy.dexy_filter import DexyFilter
from idiopidae.runtime import Composer
from ordereddict import OrderedDict
from pygments.formatters import get_formatter_for_filename
from pygments.lexers import get_lexer_by_name
from pygments.lexers import get_lexer_for_filename
from pygments.lexers.agile import PythonConsoleLexer
from pygments.lexers.agile import RubyConsoleLexer
from pygments.lexers.web import JavascriptLexer
from pygments.lexers.web import PhpLexer
import idiopidae.parser
import re

class IdioFilter(DexyFilter):
    """
    Apply idiopidae to split document into sections at ### @export
    "section-name" comments.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".html", ".tex", ".txt"]
    ALIASES = ['idio', 'idiopidae']

    def process_text_to_dict(self, input_text):
        composer = Composer()
        builder = idiopidae.parser.parse('Document', input_text + "\n\0")

        ext = self.artifact.input_ext
        name = "input_text%s" % ext

        if self.artifact.args.has_key('pyg-lexer'):
            lexer = get_lexer_by_name(self.artifact.args['pyg-lexer'])
        elif ext == '.pycon':
            lexer = PythonConsoleLexer()
        elif ext == '.rbcon':
            lexer = RubyConsoleLexer()
        elif ext in ('.json', '.dexy'):
            lexer = JavascriptLexer()
        elif ext in ('.php'):
            # If we are using idio, then our code will be in sections so we
            # need to start inline with PHP. To avoid this, use pyg instead of
            # idio. (Eventually should be able to specify lexer + options in config.)
            lexer = PhpLexer(startinline=True)
        else:
            lexer = get_lexer_for_filename(name)

        if self.artifact.args.has_key('idio'):
            idio_args = self.artifact.args['idio']
        elif self.artifact.args.has_key('pygments'):
            idio_args = self.artifact.args['pygments']
        else:
            idio_args = {}

        formatter_args = {'lineanchors' : self.artifact.web_safe_document_key() }
        formatter_args.update(idio_args)

        formatter = get_formatter_for_filename(self.artifact.filename(),
            **formatter_args)

        output_dict = OrderedDict()
        lineno = 1

        for i, s in enumerate(builder.sections):
            lines = builder.statements[i]['lines']
            if len(lines) == 0:
                next
            if not re.match("^\d+$", s):
                # Manually named section, the sectioning comment takes up a
                # line, so account for this to keep line nos in sync.
                lineno += 1
            formatter.linenostart = lineno
            formatted_lines = composer.format(lines, lexer, formatter)
            output_dict[s] = formatted_lines
            lineno += len(lines)

        return output_dict



from dexy.artifacts.file_system_json_artifact import FileSystemJsonArtifact
from dexy.topsort import CycleError
from dexy.constants import NullHandler
from dexy.controller import Controller
from dexy.document import Document
from dexy.tests.utils import tempdir
from dexy.tests.utils import divert_stdout
from modargs import args as modargs
import dexy.commands
import dexy.database
import dexy.filters.python_filters
import os

SIMPLE_PY_CONFIG = {
   "." : {
       "@simple.py|py" : {
           "contents" : "x=6\\ny=7\\nprint x*y"
        }
    }
}

CIRCULAR_CONFIG = {
   "." : {
       "@abc" : { "inputs" : ["@ghi"] },
       "@def" : { "inputs" : ["@abc"] },
       "@ghi" : { "inputs" : ["@def"] }
    }
}

NO_FILTERS_CONFIG = {
   "." : {
       "@hello.txt" : {
           "contents" : "well, hello there!"
        }
    }
}

def test_init():
    c = Controller()
    assert isinstance(c.args, dict)
    assert isinstance(c.config, dict)
    assert len(c.args) == 0
    assert len(c.config) == 0

    assert len(c.log.handlers) == 1
    # Because we didn't pass a logfile or logsdir...
    assert isinstance(c.log.handlers[0], NullHandler)

    assert len(c.reports_dirs) > 1
    assert len(c.artifact_classes) > 0

def test_init_with_artifact_class():
    args = { "artifactclass" : "FileSystemJsonArtifact" }
    c = Controller(args)
    assert c.artifact_class == FileSystemJsonArtifact

def test_init_with_db():
    args = { "dbclass" : "SqliteDatabase", "dbfile" : None, "logsdir" : None }
    c = Controller(args)
    assert isinstance(c.db, dexy.database.Database)

def test_run():
    with tempdir():
        fn = modargs.function_for(dexy.commands, "dexy")
        args = modargs.determine_kwargs(fn)
        args['globals'] = []
        os.mkdir(args['logsdir'])
        c = Controller(args)
        c.config = SIMPLE_PY_CONFIG
        c.process_config()
        assert c.members.has_key("simple.py|py")
        assert isinstance(c.members["simple.py|py"], Document)

def test_docs_with_no_filters():
    with tempdir():
        fn = modargs.function_for(dexy.commands, "dexy")
        args = modargs.determine_kwargs(fn)
        args['globals'] = []
        os.mkdir(args['logsdir'])
        c = Controller(args)
        c.config = NO_FILTERS_CONFIG
        c.process_config()
        assert c.members.has_key("hello.txt")
        assert isinstance(c.members["hello.txt"], Document)
        assert sorted(c.batch_info().keys()) == [
                "args",
                "config",
                "docs",
                "elapsed",
                "finish_time",
                "id",
                "start_time",
                "timing"
                ]

def test_circular_dependencies():
    with tempdir():
        fn = modargs.function_for(dexy.commands, "dexy")
        args = modargs.determine_kwargs(fn)
        args['globals'] = []
        os.mkdir(args['logsdir'])
        args['danger'] = True
        c = Controller(args)
        c.config = CIRCULAR_CONFIG
        with divert_stdout() as stdout:
            try:
                c.process_config()
                assert False
            except CycleError:
                assert True
            stdout_text = stdout.getvalue()
        assert "abc depends on ghi" in stdout_text
        assert "def depends on abc" in stdout_text
        assert "ghi depends on def" in stdout_text




from dexy.dexy_filter import DexyFilter
import re
import subprocess

class CucumberFilter(DexyFilter):
    """Run cucumber features."""
    INPUT_EXTENSIONS = [".feature"]
    OUTPUT_EXTENSIONS = [".html", ".txt"]
    ALIASES = ["cuke"]
    EXECUTABLE = "cucumber"
    VERSION_COMMAND = "cucumber --version"

    def process(self):
        keys = self.artifact.inputs().keys()
        rb_file = self.artifact.name.replace(".feature", "\.rb")
        matches = [k for k in keys if re.match(re.compile("^%s" % rb_file), k)]

        if len(matches) == 0:
            err_msg = "no file matching %s was found in %s" % (rb_file, keys)
            raise Exception(err_msg)
        if len(matches) > 1:
            err_msg = "too many files matching %s were found in %s: %s" % (rb_file, keys, matches)
            raise Exception(err_msg)

        key = matches[0]
        rb_art = self.artifact.inputs()[key]
        rf = rb_art.filename()
        self.artifact.generate_workfile()
        wf = self.artifact.work_filename()
        command = "%s -r %s %s" % (self.executable(), rf, wf)
        self.log.debug(command)
        env=None
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                env=env)
        stdout, stderr = proc.communicate()

        self.artifact.data_dict['1'] = stdout




from dexy.reporter import Reporter
import datetime
import os
import shutil
import tarfile

class SourceReporter(Reporter):
    """
    Saves dexy-processed files in a directory under their canonical filenames.
    Tries to be smart about which files you actually care about and want to
    see. See LongOutputReporter if you want to see everything (with uglier names).
    """

    def run(self):
        reports_dir = os.path.join(self.logsdir, "batch-source-%05d" % self.batch_id)
        self.create_reports_dir(reports_dir)

        self.load_batch_artifacts()
        for artifact in self.artifacts.values():
            if artifact.initial:
                fp = os.path.join(reports_dir, artifact.name)
                artifact.write_to_file(fp)

        print "source files saved in", reports_dir



from dexy.topsort import topsort
from dexy.topsort import CycleError

def test_topological_sort_valid():
    assert topsort([(1,2),(1,3),(3,2)]) == [1,3,2]
    assert topsort([(2,1),(2,1)]) == [2,1]
    assert topsort([(1,0),(1,2),(1,3),(3,2)]) == [1,0,3,2]

def test_topological_sort_circular_references():
    try:
        topsort([(1,2),(2,1)])
        assert False
    except CycleError as e:
        print e
        assert True

    try:
        topsort([(0,1),(1,2),(2,1)])
        assert False
    except CycleError as e:
        print e
        assert True



from dexy.constants import Constants
from dexy.topsort import CycleError
from dexy.topsort import topsort
from dexy.utils import get_log
from dexy.utils import save_batch_info
from ordereddict import OrderedDict
import copy
import dexy
import dexy.document
import dexy.introspect
import fnmatch
import glob
import json
import os
import re
import sre_constants
import time

class Controller(object):
    def __init__(self, args={}):
        self.args = args # arguments from command line
        self.config = {} # config to be processed from .dexy files
        self.docs = []
        self.timing = []

        self.batch_start_time = None
        self.batch_finish_time = None
        self.batch_elapsed_time = None

        # Set up logging
        if args.has_key("logsdir") and args.has_key("logfile"):
            self.log = get_log("dexy.controller", args['logsdir'], args['logfile'])
        else:
            self.log = Constants.NULL_LOGGER

        # Set up db
        if args.has_key('dbclass') and args.has_key("logsdir") and args.has_key("dbfile"):
            self.db = dexy.utils.get_db(self.args['dbclass'], logsdir=self.args['logsdir'], dbfile=args['dbfile'])
        else:
            self.db = None

        # List of directories that reporters use, these will not be processed by dexy
        self.reports_dirs = dexy.introspect.reports_dirs(self.log)

        # list of artifact classes - if nothing else uses this then move
        # it into the if statement below and don't cache it

        self.artifact_classes = dexy.introspect.artifact_classes(self.log)
        if args.has_key('artifactclass'):
            if self.artifact_classes.has_key(args['artifactclass']):
                self.artifact_class = self.artifact_classes[args['artifactclass']]
            else:
                raise Exception("Artifact class name %s not found in %s" % (args['artifactclass'], ",".join(self.artifact_classes.keys())))

    def run(self):
        """
        This does all the work.
        """
        self.batch_start_time = time.time()
        start = self.batch_start_time

        self.log.debug("populating Document class filter list")
        dexy.document.Document.filter_list = dexy.introspect.filters(self.log)
        self.timing.append(("populate-filter-list", time.time() - start))
        start = time.time()

        self.log.debug("loading config...")
        self.load_config()
        self.log.debug("finished loading config.")
        self.timing.append(("load-config", time.time() - start))
        start = time.time()

        self.log.debug("processing config, populating document list...")
        self.process_config()
        self.log.debug("finished processing config.")
        self.timing.append(("process-config", time.time() - start))
        start = time.time()

        if not self.args['dryrun']:
            self.docs = [doc.run() for doc in self.docs]
        self.timing.append(("run-docs", time.time() - start))

        self.batch_finish_time = time.time()
        self.batch_elapsed_time = self.batch_finish_time - self.batch_start_time

        self.log.debug("persisting batch info...")
        self.persist()
        self.log.debug("finished persisting.")
        self.log.debug("finished processing. elapsed time %s" % self.batch_elapsed_time)

    def persist(self):
        """
        Persists the database. Saves some information about this batch in a
        JSON file (for use by reporters or for debugging).
        """
        self.db.persist()
        save_batch_info(self.batch_id, self.batch_info(), self.args['logsdir'])

    def batch_info(self):
        """
        Dict of info to save
        """
        return {
            "id" : self.batch_id,
            "config" : self.config,
            "args" : self.args,
            "docs" : dict((doc.key(), doc.document_info()) for doc in self.docs),
            "start_time" : self.batch_start_time,
            "finish_time" : self.batch_finish_time,
            "elapsed" : self.batch_elapsed_time,
            "timing" : self.timing
            }

    def config_for_directory(self, path):
        """
        Determine the config applicable within a directory by looking in every
        parent directory (up as far as the dexy project root) for config files
        and combining them, such that subdirectories override parents.
        """
        global_args = {}
        config_dict = {}
        config_file = self.args['config']

        path_elements = path.split(os.sep)

        for i in range(0,len(path_elements)+1):
            config_path = os.path.join(*(path_elements[0:i] + [config_file]))
            config_files = glob.glob(config_path)

            for f in config_files:
                self.log.info("loading config file %s" % f)

                with open(f, "r") as cf:
                    try:
                        json_dict = json.load(cf)
                    except ValueError as e:
                        msg = "Your config file %s has invalid JSON\n%s" % (f, e.message)
                        raise Exception(msg)

                if json_dict.has_key("$reset"):
                    # Reset the config, i.e. ignore everything from parent
                    # directories, just use this directory's config in json_dict
                    config_dict = json_dict
                else:
                    for k in config_dict.keys():
                        propagate_virtual = config_dict[k].has_key('propagate') and config_dict[k]['propagate']
                        if k.startswith("@") and not json_dict.has_key(k) and not propagate_virtual:
                            # don't propagate virtual files
                            del config_dict[k]

                    # Combine any config in this dir with parent dir config.
                    config_dict.update(json_dict)

                if json_dict.has_key("$globals"):
                    global_args.update(json_dict["$globals"])

        config_dict['$globals'] = global_args
        return config_dict

    def load_config(self):
        """
        This method determines which subdirectories will be included in the
        dexy batch and populates the config dict for each of them.
        """
        if self.args['recurse']:

            # Figure out which directories need to be skipped
            exclude_at_root = Constants.EXCLUDE_DIRS_ROOT + self.reports_dirs + [self.args['artifactsdir'], self.args['logsdir']]
            self.log.debug("project root excluded directories %s" % ", ".join(exclude_at_root))

            exclude_everywhere = Constants.EXCLUDE_DIRS_ALL_LEVELS
            self.log.debug("directories excluded at all levels %s" % ", ".join(exclude_everywhere))

            for dirpath, dirnames, filenames in os.walk(self.args['directory']):
                # Figure out if we should process this directory and recurse
                # into its children. Start with process_dir = True
                process_dir = True

                # Remove any children we don't want to recurse into.
                if dirpath == ".":
                    for x in exclude_at_root:
                        if x in dirnames:
                            dirnames.remove(x)
                for x in exclude_everywhere:
                    if x in dirnames:
                        dirnames.remove(x)

                # Look for a .nodexy file
                if os.path.isfile(os.path.join(dirpath, '.nodexy')):
                    # If we find one...
                    self.log.info(".nodexy file found in %s" % dirpath)

                    # ...remove all child dirs from processing...
                    for i in xrange(len(dirnames)):
                        dirnames.pop()

                    # ...and skip this directory.
                    process_dir = False

                # Check if we match any excludes specified on the command line
                args_exclude = self.args['exclude']
                if isinstance(args_exclude, str):
                    args_exclude = args_exclude.split()
                for pattern in args_exclude:
                    for d in dirnames:
                        m1 = re.match(pattern, d)
                        m2 = re.match("./%s" % pattern, d)
                        m3 = re.match("%s/" % pattern, d)
                        m4 = re.match("./%s/" % pattern, d)
                        if m1 or m2 or m3 or m4:
                            dirnames.remove(d)

                if process_dir:
                    self.config[dirpath] = self.config_for_directory(dirpath)
            else:
                # Not recursing
                dirpath = self.args['directory']
                self.config[dirpath] = self.config_for_directory(dirpath)

    def process_config(self):
        """
        Processes a populated config dict, identifies files to be processed,
        creates Document objects for each, links dependencies and finally does
        topological sort to establish order of batch run.
        """

        # Define the parse_doc nested function which we will call recursively.
        def parse_doc(path, input_directive, args = {}):
            # If a specification is nested in a dependency, then input_directive
            # may be a dict. If so, split it into parts before continuing.
            try:
                a, b = input_directive.popitem()
                input_directive = a
                args = b
            except AttributeError:
                pass

            tokens = input_directive.split("|")
            if "/" in tokens[0]:
                raise Exception("paths not allowed in tokens: %s" % tokens[0])
            if path == '.':
                glob_string = tokens[0]
            else:
                glob_string = os.path.join(re.sub("^\./", "", path), tokens[0])
            filters = tokens[1:]

            docs = []

            # virtual document
            if re.search("@", glob_string):
                virtual = True

                dangerous = not args.has_key('contents')
                if dangerous and not self.args['danger']:
                    raise Exception("""
                    You are attempting to access a remote file %s.
                    You must enable --danger flag to do this.""" % glob_string)
                glob_string = glob_string.replace("@", "")
            else:
                virtual = False

            regex = fnmatch.translate(glob_string).replace(".*", "(.*)")
            matcher = re.compile(regex)

            files = glob.glob(glob_string)

            nofiles = len(files) == 0

            if nofiles and virtual:
                files = [glob_string]

            for f in files:
                create = True
                if not virtual:
                    if os.path.isdir(f):
                        create = False

                if args.has_key('disabled'):
                    if args['disabled']:
                        create = False
                        self.log.warn("document %s|%s disabled" % (f, "|".join(filters)))

                inputs = []
                if args.has_key('inputs'):
                    if isinstance(args['inputs'], str):
                        raise Exception("inputs for %s should be an array" % f)
                    for i in args['inputs']:
                        # Create document objects for input patterns (just in this directory)
                        for doc in parse_doc(path, i):
                            inputs.append(doc.key())


                m = matcher.match(f)
                if m and len(m.groups()) > 0:
                    rootname = matcher.match(f).group(1)

                # The 'ifinput' directive says that if an input exists matching
                # the specified pattern, we should create this document and it
                # will depend on the specified input.
                if args.has_key('ifinput'):
                    if isinstance(args['ifinput'], str) or isinstance(args['ifinput'], unicode):
                        ifinputs = [args['ifinput']]
                    else:
                        self.log.debug("treating input %s as iterable. class: %s" % (
                            args['ifinput'], args['ifinput'].__class__.__name__))
                        ifinputs = args['ifinput']

                    for s in ifinputs:
                        self.log.debug("evaluating ifinput %s" % s)
                        ifinput = s.replace("%", rootname)
                        self.log.debug("evaluating ifinput %s" % ifinput)
                        input_docs = parse_doc(path, ifinput, {})
                        for input_doc in input_docs:
                            inputs.append(input_doc.key())

                    if len(input_docs) == 0:
                        create = False

                if args.has_key('ifnoinput'):
                    ifinput = args['ifnoinput'].replace("%", rootname)
                    input_docs = parse_doc(path, ifinput, {})

                    if len(input_docs) > 0:
                        create = False

                if args.has_key('except'):
                    try:
                        except_re = re.compile(args['except'])
                    except sre_constants.error as e:
                        raise Exception("""You passed 'except' value of %s.
Please pass a valid Python-style regular expression for
'except', NOT a glob-style matcher. Error message from
re.compile: %s""" % (args['except'], e))
                    if re.match(except_re, f):
                        self.log.warn("skipping %s for %s as it matches except pattern %s" % (
                                f,
                                input_directive,
                                args['except']
                                ))
                        create = False

                if create:
                    doc = dexy.document.Document()
                    doc.set_controller(self)

                    # Filters can either be included in the name...
                    doc.set_name_and_filters(f, filters)
                    # ...or they may be listed explicitly.
                    if args.has_key('filters'):
                        doc.filters += args['filters']

                    doc.setup_log() # After name has been set
                    doc.virtual = virtual

                    key = doc.key()
                    self.log.debug("creating doc %s for glob %s" % (key, glob_string))

                    if self.members.has_key(key):
                        self.log.debug("found existing key %s" % key)
                        doc = self.members[key]
                    else:
                        self.log.debug("no existing key %s" % key)

                    if args.has_key('priority'):
                        doc.priority = args['priority']
                        del args['priority']

                    doc.args.update(args)

                    if args.has_key('allinputs'):
                        doc.use_all_inputs = args['allinputs']

                    if args.has_key('inputs'):
                        doc.input_args = copy.copy(args['inputs'])
                        doc.input_keys = []

                    for i in inputs:
                        doc.add_input_key(i)

                    self.members[key] = doc
                    docs.append(doc) # docs is a local list of docs

            return docs # end of parse_doc nested function

        def get_pos(member):
            key = member.key()
            return self.members.keys().index(key)

        def depend(parent, child):
            self.depends.append((get_pos(child), get_pos(parent)))

        # The real processing starts here.
        self.members = OrderedDict()
        self.depends = []

        self.batch_id = self.db.next_batch_id()
        print "batch id is", self.batch_id

        for path, config in self.config.iteritems():
            ### @export "features-global-args-1"
            if config.has_key("$globals"):
                global_args = config["$globals"]
            else:
                global_args = {}

            if self.args.has_key('globals'):
                global_args.update(self.args['globals'])

            for k, v in config.iteritems():
                local_args = global_args.copy()
                local_args.update(v)
                for kg in global_args.keys():
                    if local_args.has_key(kg):
                        if isinstance(local_args[kg], dict):
                            local_args[kg].update(global_args[kg])
                parse_doc(path, k, local_args)
            ### @end

        # Determine dependencies
        total_dependencies = 0
        for doc in self.members.values():
            doc.finalize_inputs(self.members)
            total_dependencies += len(doc.inputs)
            for input_doc in doc.inputs:
                depend(doc, input_doc)

        if len(self.args['run']) > 0:
            # Only run the specified document, and its dependencies.
            new_members = OrderedDict()
            new_depends = []

            def new_get_pos(member):
                key = member.key()
                return new_members.keys().index(key)

            def new_depend(parent, child):
                new_depends.append((new_get_pos(child), new_get_pos(parent)))

            def parse_new_document(d):
                new_members[d.key()] = d
                for input_doc in d.inputs:
                    if not input_doc.key() in new_members.keys():
                        new_members[input_doc.key()] = input_doc
                    new_depend(d, input_doc)
                    parse_new_document(input_doc)

            run_key = self.args['run']
            doc = self.members[run_key]
            parse_new_document(doc)

            print "limiting members list to %s and its dependencies, %s/%s documents will be run" % (doc.key(), len(new_members), len(self.members))
            self.members = new_members
            self.depends = new_depends

        num_members = len(self.members)
        if num_members > 0:
            dep_ratio = float(total_dependencies)/num_members
        else:
            dep_ratio = None

        print "sorting %s documents into run order, there are %s total dependencies" % (num_members, total_dependencies)
        if dep_ratio:
            print "ratio of dependencies to documents is %0.1f" % (dep_ratio)
            if dep_ratio > 10:
                print "if you are experiencing performance problems:"
                print "call dexy with -dryrun and inspect logs/batch-XXXX.json to debug dependencies"
                print "consider using -strictinherit or reducing your use of 'allinputs' "

        try:
            self.log.debug("Beginning topological sort...")
            topsort_ordering = topsort(self.depends)
            self.log.debug("Topological sort completed successfully.")
        except CycleError as e:
            print "There are circular dependencies!"
            answer, num_parents, children = e.args
            for child, parents in children.items():
                for parent in parents:
                    print "%s depends on %s" % (self.members.keys()[parent], self.members.keys()[child])
            raise e

        docs_without_dependencies = frozenset(range(len(self.members))) - frozenset(topsort_ordering)
        self.ordering = topsort_ordering + list(docs_without_dependencies)

        for i in self.ordering:
            key = self.members.keys()[i]
            self.docs.append(self.members[key])



from dexy.dexy_filter import DexyFilter
from dexy.document import Document
import dexy.filters.pygments_filters
import dexy.introspect

def test_get_filter_for_alias():
    Document.filter_list = dexy.introspect.filters()
    d = Document()
    for k, v in d.__class__.filter_list.iteritems():
        assert isinstance(k, str)
        assert issubclass(v, DexyFilter)

    assert d.get_filter_for_alias("pyg") == dexy.filters.pygments_filters.PygmentsFilter
    assert d.get_filter_for_alias("-") == DexyFilter
    assert d.get_filter_for_alias("-xxxalias") == DexyFilter

def test_document_set_name_and_filters():
    doc = Document()
    doc.set_name_and_filters("data/test.py|abc")
    assert doc.name == "data/test.py"
    assert doc.filters == ['abc']

    doc.filters += ['def', 'xyz']
    assert doc.filters == ['abc', 'def', 'xyz']

    assert doc.key() == "data/test.py|abc|def|xyz"



from dexy.constants import Constants
from dexy.controller import Controller
from dexy.utils import get_log
from dexy.version import Version
from modargs import args
from pygments import highlight
from pygments.lexers.agile import PythonLexer
from pygments.formatters import TerminalFormatter
import cProfile
import copy
import datetime
import dexy.introspect
import inspect
import os
import shutil
import sys

MOD = sys.modules[__name__]
PROG = 'dexy'
S = "   "

# List of filter classes not to include in the filters command
NODOC_FILTERS = [
    "SubprocessCompileFilter",
    "SubprocessCompileInputFilter",
    "SubprocessFilter",
    "SubprocessStdoutFilter",
    "SubprocessStdoutInputFileFilter",
    "SubprocessStdoutInputFilter"
    ]

def run():
    args.parse_and_run_command(sys.argv[1:], MOD, default_command=Constants.DEFAULT_COMMAND)

def dexy_command(
        allreports=False, # whether to run all available reports
        artifactclass=Constants.DEFAULT_ACLASS, # name of class to use for artifacts
        artifactsdir=Constants.DEFAULT_ADIR, # location of directory in which to store artifacts
        config=Constants.DEFAULT_CONFIG, # name to use for configuration file
        danger=False, # whether to allow running remote files
        dbclass=Constants.DEFAULT_DBCLASS, # name of database class to use
        dbfile=Constants.DEFAULT_DBFILE, # name of the database file (it lives in the logs dir)
        directory=".", # the directory to process, you can just process a subdirectory of your project
        disabletests=False, # Whether to disable the dexy 'test' filter
        dryrun=False, # if True, just parse config and print batch info, don't run dexy
        exclude="", # directories to exclude from dexy processing
        filters=False, # DEPRECATED just to catch people who use the old dexy --filters syntax
        globals="DEXY_VERSION=%s" % Version.VERSION, # global values to make available within dexy documents, should be KEY=VALUE pairs separated by spaces
        help=False, # DEPRECATED just to catch people who use the old dexy --help syntax
        hashfunction='md5', # What hash function to use, set to crc32 or adler32 for more speed, less reliability
        ignore=False, # whether to ignore nonzero exit status or raise an error - may not be supported by all filters
        logfile=Constants.DEFAULT_LFILE, # name of log file
        logsdir=Constants.DEFAULT_LDIR, # location of directory in which to store logs
        nocache=False, # whether to force artifacts to run even if there is a matching file in the cache
        output=False, # Shortcut to mean "I just want the OutputReporter, nothing else"
        recurse=True, # whether to recurse into subdirectories when running Dexy
        reporters=False, # DEPRECATED just to catch people who use the old dexy --reporters syntax
        reports=Constants.DEFAULT_REPORTS, # reports to be run after dexy runs, enclose in quotes and separate with spaces
        reset=False, # whether to purge existing artifacts and logs before running Dexy
        run="", # specific document to run. if specified, this document + its dependencies will be all that is run
        setup=False, # DEPRECATED just to catch people who use the old dexy --setup syntax
        strictinherit=False, # set to true if you want 'allinputs' to only reference items in same dir or a subdir
        uselocals=True, # use cached local copies of remote URLs, faster but might not be up to date, 304 from server will override this setting
        version=False # DEPRECATED just to catch people who use the old dexy --version syntax
    ):
    """
    Runs Dexy, by processing your .dexy configuration file and running content
    through the filters you have specified. Results are cached in the
    artifacts/ directory but are presented in a more usable format by
    reporters. Basic reports are run automatically but you can specify
    additional reports. Type 'dexy reporters' for a list of available reporters.

    If your project is large, then running reports will start to take up a lot
    of time, so you should specify only the reports you really need. You can
    always run more reports after a batch has finished running (you can run
    historical reports as far back as the last time you cleared out your
    artifacts cache with a 'dexy reset' or similar).

    After running Dexy, the output/ directory will hold what dexy thinks are
    the most important generated files (with pretty filenames), the output-long
    directory will hold all of your generated files (with ugly filenames), and
    the logs/ directory will hold the basic dexy.log logfile and also a more
    colorful and descriptive HTML log file in logs/run-latest/. Please look at
    these logfiles to learn more about how dexy works, and if you run into
    problems the dexy.log file might provide clues as to what has gone wrong.

    Your original files will be copied to logs/source-batch-00001/ by the
    SourceReporter (enabled by default). Each time you run dexy, your source
    code files will be copied so you have a mini-version history. (You can also
    use the 'dexy history' command to get a history for a given file, and you
    can run the SourceReporter again at any time to restore a given batch's
    source files.)

    If you run into trouble, visit http://dexy.it/help
    """
    # catch deprecated arguments
    if help or version or reporters or filters:
        print "the command syntax has changed, please type 'dexy help' for help"
        sys.exit(1)

    ### @export "dexy-command-check-setup"
    if not check_setup(logsdir=logsdir, artifactsdir=artifactsdir):
        print "Please run '%s setup' first to create the directories dexy needs to work with" % PROG
        sys.exit(1)

    ### @export "dexy-command-process-args"
    if reset:
        reset_command(logsdir=logsdir, artifactsdir=artifactsdir)

    if output:
        if not reports == Constants.DEFAULT_REPORTS:
            raise Exception("if you pass --output you can't also modify reports! pick 1!")
        if allreports:
            raise Exception("if you pass --output you can't also pass --allreports!")
        reports = "Output"

    if allreports:
        if not reports == Constants.DEFAULT_REPORTS:
            raise Exception("if you pass --allreports you can't also specify --reports")

    controller = run_dexy(locals())
    if not dryrun:
        if allreports:
            reports_command(
                allreports=True,
                artifactclass=artifactclass,
                controller=controller,
                hashfunction=hashfunction,
                logsdir=logsdir
            )
        else:
            reports_command(
                reports=reports,
                artifactclass=artifactclass,
                controller=controller,
                hashfunction=hashfunction,
                logsdir=logsdir
            )


def run_dexy(args):
    # validate args and do any conversions required
    args['globals'] = dict([g.split("=") for g in args['globals'].split()])
    args['exclude'] = [x.strip("/") for x in args['exclude'].split()]
    controller = Controller(args)
    controller.run()
    return controller

def profile_command(
        reports="ProfileReporter",
        n=1, # How many times to run dexy for profiling.
        **kw # Accepts additional keyword arguments for the 'dexy' command
    ):
    """
    Runs dexy using cProfile to do time-based profiling. Uses ProfileReport
    (the only report enabled by default) to present profiling information.
    Other reports can be specified, report time is not included in profiling.
    Running ProfileReport each time ensures that profiling data is stored in
    sqlite database for comparison (a 'dexy reset' will delete this database).
    """
    dexy_fn = args.function_for(dexy.commands, "dexy")
    defaults = args.determine_kwargs(dexy_fn)
    defaults.update(kw)
    defaults['profile'] = True

    locals_for_run_dexy = {'args' : defaults}

    logs_dir = kw.has_key("logsdir") and kw['logsdir'] or Constants.DEFAULT_LDIR
    prof_file = os.path.join(logs_dir, "dexy.prof")

    report_kwargs = {}
    if kw.has_key('artifactclass'):
        report_kwargs['artifactclass'] = kw['artifactclass']

    for i in xrange(n):
        print "===== run %s of %s =====" % (i+1, n)
        cProfile.runctx("run_dexy(args)", globals(), copy.deepcopy(locals_for_run_dexy), prof_file)
        reports_command(reports=reports, **report_kwargs)

def check_setup(logsdir=Constants.DEFAULT_LDIR, artifactsdir=Constants.DEFAULT_ADIR):
    return os.path.exists(logsdir) and os.path.exists(artifactsdir)

def setup_command(logsdir=Constants.DEFAULT_LDIR, artifactsdir=Constants.DEFAULT_ADIR, logfile=Constants.DEFAULT_LFILE, showhelp=True):
    """
    Creates directories to hold artifacts and logs. Dexy needs these
    directories and they aren't created automatically as a precaution against
    accidentally running dexy somewhere you didn't mean to.
    """
    if check_setup(logsdir=logsdir, artifactsdir=artifactsdir):
        print "Dexy is already set up. Run 'reset' if you want to reset everything."
    else:
        if not os.path.exists(logsdir):
            os.mkdir(logsdir)
        if not os.path.exists(artifactsdir):
            os.mkdir(artifactsdir)

        if showhelp:
            print
            print "Ok, we've created directories called %s and %s." % (logsdir, artifactsdir)
            if os.path.exists(Constants.DEFAULT_CONFIG):
                print "You are now ready to run dexy!  If you have problems,"
                print "please check the log file at %s/%s for clues." % (logsdir, logfile)
                print "Online help is available from dexy.it/help"
            else:
                print "You are almost ready to run dexy! You just need to create a config file,"
                print "check out the tutorials dexy.it/docs/tutorials if you aren't sure how."
                print "You can type '%s help -on %s' (without quotes) for help running dexy" % (PROG, Constants.DEFAULT_COMMAND)
                print "or visit dexy.it/help for more resources"
            print

def cleanup_command(logsdir=Constants.DEFAULT_LDIR, artifactsdir=Constants.DEFAULT_ADIR):
    """
    Removes all dexy-generated content.
    """
    # Dexy itself should put all generated content into either artifacts or logs.
    # Reports may create files in other locations, but they should track where
    # they do this and be able to purge them.
    purge_artifacts(remake=False, artifactsdir=artifactsdir)
    purge_logs(logsdir=logsdir)
    purge_reports()

def history_command(
        filename=None,
        logsdir=Constants.DEFAULT_LDIR,
        artifactsdir=Constants.DEFAULT_ADIR,
        dbclass=Constants.DEFAULT_DBCLASS,
        dbfile=Constants.DEFAULT_DBFILE
        ):
    """
    Returns a list of available versions of the file. This must be run from the
    dexy project root.
    """
    db = dexy.utils.get_db(dbclass, dbfile=dbfile, logsdir=logsdir)
    versions = {}
    for row in db.all():
        key = row['key']
        batch_id = int(row['batch_id'])
        if key == filename and not versions.has_key(batch_id):
            versions[batch_id] = row

    if len(versions) > 0:
        print
        print "Dexy found these versions of %s:" % filename
        for b in sorted(versions.keys()):
            row = versions[b]
            artifact_file = os.path.join(artifactsdir, "%s%s" % (row['hashstring'], row['ext']))

            batch_source_dir = os.path.join(logsdir, "batch-source-%0.5d" % b)
            batch_source_file = os.path.join(batch_source_dir, row['key'])
            time = row['mtime']
            if time and not time == 'None':
                human_time = datetime.datetime.fromtimestamp(float(time))
            else:
                human_time = "NA"

            batch_info_file = dexy.utils.batch_info_filename(b, logsdir)
            batch_ok = os.path.exists(batch_info_file)

            artifact_ok = os.path.exists(artifact_file)
            batch_source_ok = os.path.exists(batch_source_file)

            if artifact_ok and batch_ok:
                msg = "  batch id %5d  modified time %s available in %s" % (b, human_time, artifact_file)
                if batch_source_ok:
                    msg += " and %s" % batch_source_file
                print msg
        print
    else:
        print "No versions found for", filename

def reset_command(logsdir=Constants.DEFAULT_LDIR, artifactsdir=Constants.DEFAULT_ADIR):
    """
    Runs cleanup and then setup to completely reset your dexy environment.
    """
    cleanup_command(logsdir=logsdir, artifactsdir=artifactsdir)
    setup_command(logsdir=logsdir, artifactsdir=artifactsdir, showhelp=False)

def purge_artifacts(remake=True, artifact_class=None, artifactsdir=Constants.DEFAULT_ADIR):
    shutil.rmtree(artifactsdir, ignore_errors=True)
    if remake:
        os.mkdir(artifactsdir)

    if hasattr(artifact_class, 'purge'):
        artifact_class.purge()

def purge_logs(logsdir=Constants.DEFAULT_LDIR):
    shutil.rmtree(logsdir, ignore_errors=True)

def purge_reports():
    reports_dirs = dexy.introspect.reports_dirs()
    for d in reports_dirs:
        safety_file = os.path.join(d, ".dexy-generated")
        if d and os.path.exists(d) and os.path.exists(safety_file):
            print "purging contents of %s" % d
            shutil.rmtree(d)
        elif d and os.path.exists(d):
            print "not purging %s, please remove this directory manually" % d

def help_command(on=False):
    args.help_command(PROG, MOD, Constants.DEFAULT_COMMAND, on)

def help_text(on=False):
    return args.help_text(PROG, MOD, Constants.DEFAULT_COMMAND, on)

def version_command():
    """Print the current version."""
    print "%s version %s" % (PROG, Version.VERSION)

def filters_command(
        alias="", # If a filter alias is specified, more detailed help for that filter is printed.
        nocolor=False, # When source = True, whether to omit syntax highlighting
        showall=False, # Whether to show all filters, including those which need missing software, implies versions=True
        showmissing=False, # Whether to just show filters missing external software, implies versions=True
        space=False, # Whether to add extra spacing to the output for extra readability
        source=False, # Whether to include syntax-highlighted source code when displaying an indvidual filter
        versions=False # Whether to check the installed version of external software required by filters, slower
    ):
    """
    Lists the available dexy filters and their aliases.

    Does not include filters which require python modules that are not
    installed.

    Consult the Dexy website http://dexy.it for a complete list of filters.
    """
    if len(alias) == 0:
        # This tends to be slow, let people know it's running
        print "looking up filter information..."
    else:
        if showall:
            raise Exception("can't specify an alias if showall is True")

    if check_setup():
        log = get_log()
    else:
        log = Constants.NULL_LOGGER

    print filters_text(alias, nocolor, showall, showmissing, space, source, versions, log)

def filters_text(
        alias="", # If a filter alias is specified, more detailed help for that filter is printed.
        nocolor=False, # When source = True, whether to omit syntax highlighting
        showall=False, # Whether to show all filters, including those which need missing software, implies versions=True
        showmissing=False, # Whether to just show filters missing external software, implies versions=True
        space=False, # Whether to add extra spacing to the output for extra readability
        source=False, # Whether to include syntax-highlighted source code when displaying an indvidual filter
        versions=False, # Whether to check the installed version of external software required by filters, slower
        log=Constants.NULL_LOGGER
        ):
    filters_dict = dexy.introspect.filters(log)

    if len(alias) > 0:
        # We want help on a particular filter
        klass = filters_dict[alias]
        text = []
        text.append(klass.__name__)
        text.append("")
        text.append("Aliases: %s" % ", ".join(klass.ALIASES))
        text.append("")
        text.append(trim_docstring(klass.__doc__))
        text.append("")
        text.append("http://dexy.it/docs/filters/%s" % alias)
        if source:
            text.append("")
            source_code = inspect.getsource(klass)
            if nocolor:
                text.append(source_code)
            else:
                formatter = TerminalFormatter()
                lexer = PythonLexer()
                text.append(highlight(source_code, lexer, formatter))
        return "\n".join(text)

    else:
        def sort_key(k):
            return k.__name__

        filter_classes = sorted(set(f for f in filters_dict.values()), key=sort_key) # uniqify and sort class names

        text = []
        for klass in filter_classes:
            if not showall:
                skip = klass.__name__ in NODOC_FILTERS
            else:
                skip = False

            if (versions or showmissing or showall) and not skip:
                version = klass.version()
                no_version_info_available = (version is None)
                if no_version_info_available:
                    version_message = ""
                    if showmissing:
                        skip = True
                elif version:
                    version_message = "Installed version: %s" % version
                    if showmissing:
                        skip = True
                else:
                    if not (showmissing or showall):
                        skip = True
                    version_message = "'%s' failed, filter may not be available." % klass.version_command()

            if not skip:
                name_and_aliases = "%s (%s) " % (klass.__name__, ", ".join(klass.ALIASES))
                docstring = trim_docstring(klass.__doc__)
                if "\n" in docstring:
                    docstring = docstring.splitlines()[0]
                filter_help = name_and_aliases + docstring
                if (versions or showmissing or (showall and not version)):
                    filter_help += " %s" % version_message
                text.append(filter_help)

        if space:
            sep = "\n\n"
        else:
            sep = "\n"
        return sep.join(text)

def reporters_command():
    """Lists currently available dexy reporters."""
    print reporters_text(Constants.NULL_LOGGER)

def reporters_text(log):
    text = []
    reporters = dexy.introspect.reporters(log)
    for reporter_name in sorted(reporters.keys()):
        r = reporters[reporter_name]
        text.append("\n" + r.__name__.replace("Reporter", ""))

        if r.__doc__:
            text.append(S + r.__doc__)
        else:
            text.append(S + "no documentation available")

        if r.REPORTS_DIR:
            text.append(S + "any generated reports will be saved in %s" % r.REPORTS_DIR)

    return "\n".join(text) + "\n"

def report_command(**kwargs):
    """
    Alias for reports.
    """
    if kwargs.has_key('report'):
        kwargs['reports'] = kwargs['report']
        del kwargs['report']
    reports_command(**kwargs)

def reports_command(
        allreports=False, # whether to run all available reporters (except those that are specifically excluded by setting ALLREPORTS=False)
        artifactclass=Constants.DEFAULT_ACLASS, # What artifact class to use (must correspond to artifact class used by batch id)
        batchid=False, # What batch id to run reports for. Leave false to run the most recent batch available (recommended).
        controller=False, # can pass the just-run controller instance to save having to load some data from disk which is already in memory
        hashfunction='md5',
        logsdir=Constants.DEFAULT_LDIR, # The location of the logs directory.
        reports="OutputReporter RunReporter LongOutputReporter OutputTgzReporter" # The class names for all reports to be run.
    ):
    """
    Runs reports to present Dexy output.
    """

    # convert False (needed for CLI) to None (expected by constructor)
    if not batchid:
        batchid = None
    if not controller:
        controller = None

    if isinstance(reports, str):
        reports = reports.split()
    elif isinstance(reports, bool):
        # either was set to False, or an empty string evaluated to True
        # either way we don't want to run any reports.
        reports = []
    reporters = dexy.introspect.reporters()
    if allreports:
        print "using allreports"
        reports = [k for k, v in reporters.items() if v.ALLREPORTS]

    for r in reports:
        reporter_class_name = r.endswith("Reporter") and r or ("%sReporter" % r)
        if not reporters.has_key(reporter_class_name):
            raise Exception("No reporter class named %s available. Valid reporter classes are: %s" % (r, ", ".join(reporters.keys())))
        report_class = reporters[reporter_class_name]

        print "running", r
        reporter = report_class(
                artifact_class=artifactclass,
                batch_id=batchid,
                controller=controller,
                hashfunction=hashfunction,
                logsdir=logsdir
            )

        reporter.load_batch_artifacts()
        reporter.run()

def it_command(**kwargs):
    dexy_command(kwargs)

# From http://www.python.org/dev/peps/pep-0257/#handling-docstring-indentation
def trim_docstring(docstring):
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxint
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxint:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)



from dexy.filters.process_filters import SubprocessStdoutFilter
from dexy.filters.process_filters import SubprocessStdoutInputFilter
from dexy.filters.process_filters import SubprocessStdoutInputFileFilter

class RhinoSubprocessStdoutFilter(SubprocessStdoutFilter):
    EXECUTABLE = "rhino -f"
    INPUT_EXTENSIONS = [".js"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['js', 'rhino']

class CowsaySubprocessStdoutFilter(SubprocessStdoutFilter):
    ALIASES = ['cowsay']
    EXECUTABLE = 'cowsay'
    INPUT_EXTENSIONS = [".txt"]
    OUTPUT_EXTENSIONS = [".txt"]

    def command_string_stdout(self):
        args = self.command_line_args() or ""
        text = self.artifact.input_text()
        return "%s %s \"%s\"" % (self.executable(), args, text)

class CowthinkSubprocessStdoutFilter(CowsaySubprocessStdoutFilter):
    ALIASES = ['cowthink']
    EXECUTABLE = 'cowthink'

class PythonSubprocessStdoutFilter(SubprocessStdoutFilter):
    ALIASES = ['py', 'pyout']
    EXECUTABLE = 'python'
    INPUT_EXTENSIONS = [".py", ".txt"]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = 'python --version'

class PythonSubprocessStdoutInputFilter(SubprocessStdoutInputFilter):
    ALIASES = ['pyinput']
    EXECUTABLE = 'python'
    INPUT_EXTENSIONS = [".py", ".txt"]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = 'python --version'

class BashSubprocessStdoutFilter(SubprocessStdoutFilter):
    ALIASES = ['sh', 'bash']
    EXECUTABLE = 'bash'
    INPUT_EXTENSIONS = [".sh", ".bash", ".txt", ""]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = 'bash --version'

class BashSubprocessStdoutInputFilter(SubprocessStdoutInputFilter):
    ALIASES = ['shinput']
    EXECUTABLE = 'bash'
    INPUT_EXTENSIONS = [".sh", ".bash", ".txt", ""]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = 'bash --version'

class SedSubprocessStdoutInputFilter(SubprocessStdoutInputFilter):
    ALIASES = ['sed']
    EXECUTABLE = 'sed'
    INPUT_EXTENSIONS = [".sed"]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = 'sed --version'

    def command_string_stdout(self):
        wf = self.artifact.previous_artifact_filename
        return "%s -f %s" % (self.executable(), wf)

class RegetronSubprocessStdoutInputFileFilter(SubprocessStdoutInputFileFilter):
    ALIASES = ['regetron']
    EXECUTABLE = 'regetron'
    INPUT_EXTENSIONS = [".regex"]
    OUTPUT_EXTENSIONS = [".txt"]

    def command_string_stdout_input(self, input_artifact):
        wf = self.artifact.previous_artifact_filename
        input_file = input_artifact.filename()
        return "%s %s %s" % (self.executable(), input_file, wf)

class IrbSubprocessStdoutFilter(SubprocessStdoutFilter):
    ALIASES = ['irbout']
    EXECUTABLE = 'irb --simple-prompt --noreadline'
    INPUT_EXTENSIONS = [".txt", ".rb"]
    OUTPUT_EXTENSIONS = [".rbcon"]
    VERSION_COMMAND = 'irb --version'

class IrbSubprocessStdoutInputFilter(SubprocessStdoutInputFilter):
    ALIASES = ['irboutinput']
    EXECUTABLE = 'irb --simple-prompt --noreadline'
    INPUT_EXTENSIONS = [".txt", ".rb"]
    OUTPUT_EXTENSIONS = [".rbcon"]
    VERSION_COMMAND = 'irb --version'

class RubySubprocessStdoutFilter(SubprocessStdoutFilter):
    EXECUTABLE = 'ruby'
    VERSION_COMMAND = 'ruby --version'
    INPUT_EXTENSIONS = [".txt", ".rb"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['rb']

class RubySubprocessStdoutInputFilter(SubprocessStdoutInputFilter):
    EXECUTABLE = 'ruby'
    VERSION_COMMAND = 'ruby --version'
    INPUT_EXTENSIONS = [".txt", ".rb"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['rbinput']

class EscriptSubprocessStdoutFilter(SubprocessStdoutFilter):
    """
    Runs Erlang scripts using the escript command.
    """
    EXECUTABLE = 'escript'
    INPUT_EXTENSIONS = [".erl"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['escript']

# Unchecked below this...

class PhpFilter(SubprocessStdoutFilter):
    """
    Runs php script and returns STDOUT.
    """
    EXECUTABLE = 'php'
    VERSION_COMMAND = 'php --version'
    INPUT_EXTENSIONS = [".php"]
    OUTPUT_EXTENSIONS = [".html", ".txt"]
    ALIASES = ['php']

class LuaFilter(SubprocessStdoutFilter):
    """
    Runs lua script and returns STDOUT.
    """
    EXECUTABLE = 'lua'
    VERSION_COMMAND = 'lua -v'
    INPUT_EXTENSIONS = ['.lua']
    OUTPUT_EXTENSIONS = ['.txt']
    ALIASES = ['lua']

class RedclothFilter(SubprocessStdoutFilter):
    """
    Runs redcloth, converts textile markup to HTML.
    """
    EXECUTABLE = 'redcloth'
    VERSION_COMMAND = 'redcloth --version'
    INPUT_EXTENSIONS = [".txt", ".textile"]
    OUTPUT_EXTENSIONS = [".html"]
    ALIASES = ['redcloth', 'textile']

class RedclothLatexFilter(SubprocessStdoutFilter):
    """
    Runs redcloth, converts textile markup to LaTeX.
    """
    EXECUTABLE = 'redcloth -o latex'
    VERSION_COMMAND = 'redcloth --version'
    INPUT_EXTENSIONS = [".txt", ".textile"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['redclothl', 'latextile']

class Rst2HtmlFilter(SubprocessStdoutFilter):
    EXECUTABLE = 'rst2html.py'
    VERSION_COMMAND = 'rst2html.py --version'
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".html"]
    ALIASES = ['rst2html']
    FINAL = True

class Rst2LatexFilter(SubprocessStdoutFilter):
    EXECUTABLE = 'rst2latex.py'
    VERSION_COMMAND = 'rst2latex.py --version'
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['rst2latex']

class Rst2BeamerFilter(SubprocessStdoutFilter):
    EXECUTABLE = 'rst2beamer'
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['rst2beamer']

class SloccountFilter(SubprocessStdoutFilter):
    EXECUTABLE = 'sloccount'
    VERSION_COMMAND = 'sloccount --version'
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['sloc', 'sloccount']

class RdConvFilter(SubprocessStdoutFilter):
    """Convert R documentation to other formats."""
    EXECUTABLE = "R CMD Rdconv"
    VERSION_COMMAND = "R CMD Rdconv -v"
    INPUT_EXTENSIONS = ['.Rd']
    OUTPUT_EXTENSIONS = ['.txt', '.html', '.tex', '.R']
    ALIASES = ['rdconv']
    EXTENSION_TO_FORMAT = {
        '.txt' : 'txt',
        '.html' : 'html',
        '.tex' : 'latex',
        '.R' : 'example'
    }

    def instance_executable(self):
        return "%s --type=%s" % (self.EXECUTABLE, self.EXTENSION_TO_FORMAT[self.artifact.ext])

class RagelRubyDotFilter(SubprocessStdoutFilter):
    """
    Generates state chart in .dot format of ragel state machine for ruby.
    """
    INPUT_EXTENSIONS = [".rl"]
    OUTPUT_EXTENSIONS = [".dot"]
    ALIASES = ['rlrbd', 'ragelrubydot']
    VERSION_COMMAND = 'ragel --version'
    EXECUTABLE = 'ragel -R -V'

class LynxDumpFilter(SubprocessStdoutFilter):
    """
    Converts HTML to plain text by using lynx -dump.
    """
    INPUT_EXTENSIONS = [".html"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['lynxdump']
    VERSION_COMMAND = 'lynx --version'
    EXECUTABLE = 'lynx -dump'

class NonexistentFilter(SubprocessStdoutFilter):
    ALIASES = ['zzzdoesnotexist']
    VERSION_COMMAND = 'notherenopexxx --version'



from dexy.dexy_filter import DexyFilter
from dexy.filters.pexpect_filters import PexpectReplFilter
from dexy.filters.process_filters import SubprocessCompileFilter
from dexy.filters.process_filters import SubprocessStdoutFilter
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.formatters.latex import LatexFormatter
from pygments.lexers.compiled import JavaLexer
import json
import os
import platform
import shutil

class JrubyFilter(SubprocessStdoutFilter):
    ALIASES = ['jruby']
    EXECUTABLE = "jruby"
    INPUT_EXTENSIONS = [".rb"]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = "jruby --version"

class JirbFilter(PexpectReplFilter):
    ALIASES = ['jirb']
    ALLOW_MATCH_PROMPT_WITHOUT_NEWLINE = True
    EXECUTABLE = "jirb --prompt-mode simple"
    IGNORE_ERRORS = True # Returns nonzero exit code even when no errors.
    INPUT_EXTENSIONS = [".rb"]
    OUTPUT_EXTENSIONS = [".rbcon"]
    PROMPTS = ['>>', '?>']
    VERSION_COMMAND = "jirb --version"

class JythonFilter(SubprocessStdoutFilter):
    ALIASES = ['jython']
    EXECUTABLE = "jython"
    INPUT_EXTENSIONS = [".py"]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = "jython --version"

    @classmethod
    def enabled(self):
        if platform.system() in ('Linux', 'Windows'):
            return True
        elif platform.system() in ('Darwin'):
            if hasattr(self, 'log'):
                self.log.warn("The jython dexy filter should not be run on MacOS due to a serious bug. This filter is being disabled.")
            return False
        else:
            if hasattr(self, 'log'):
                self.log.warn("""Can't detect your system. If you see this message please report this to the dexy project maintainer, your platform.system() value is '%s'. The jython dexy filter should not be run on MacOS due to a serious bug.""" % platform.system())
            return True

class JythonInteractiveFilter(PexpectReplFilter):
    ALIASES = ['jythoni']
    EXECUTABLE = "jython -i"
    INPUT_EXTENSIONS = [".py", ".txt"]
    OUTPUT_EXTENSIONS = [".pycon"]
    VERSION_COMMAND = "jython --version"

    @classmethod
    def enabled(self):
        if platform.system() in ('Linux', 'Windows'):
            return True
        elif platform.system() in ('Darwin'):
            print "The jythoni dexy filter should not be run on MacOS due to a serious bug. This filter is being disabled."
            return False
        else:
            print """Can't detect your system. If you see this message please report this to the dexy project maintainer, your platform.system() value is '%s'. The jythoni dexy filter should not be run on MacOS due to a serious bug.""" % platform.system()
            return True

class JavaFilter(SubprocessCompileFilter):
    ALIASES = ['java']
    CHECK_RETURN_CODE = True # Whether to check return code when running compiled executable.
    COMPILED_EXTENSION = ".class"
    EXECUTABLE = "javac"
    FINAL = False
    INPUT_EXTENSIONS = [".java"]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = "java -version"

    def setup_cp(self):
        """
        Makes sure the current working directory is on the classpath, also adds
        any specified CLASSPATH elements. Assumes that CLASSPATH elements are either
        absolute paths, or paths relative to the artifacts directory. Also, if
        an input has been passed through the javac filter, its directory is
        added to the classpath.
        """
        env = self.setup_env()

        classpath_elements = []
        classpath_elements.append(os.path.dirname(self.artifact.name))

        for key, input_artifact in self.artifact.inputs().iteritems():
            if input_artifact.ext == ".class" and "javac" in key:
                classpath_elements.append(os.path.dirname(input_artifact.name))

        if env and env.has_key("CLASSPATH"):
            for x in env['CLASSPATH'].split(":"):
                if x.startswith("/"):
                    # absolute path, leave it alone
                    classpath_elements.append(x)
                else:
                    # path relative to artifacts directory, need to adjust
                    # since we are working in a subdir of artifacts
                    classpath_elements.append(os.path.join("..", x))

        cp = ":".join(classpath_elements)

        return cp

    def compile_command_string(self):
        cp = self.setup_cp()
        return "javac -classpath %s %s" % (cp, self.artifact.name)

    def run_command_string(self):
        cp = self.setup_cp()
        main_method = self.setup_main_method()
        args = self.command_line_args() or ""
        return "java %s -cp %s %s" % (args, cp, main_method)

    def setup_cwd(self):
        tempdir = self.artifact.temp_dir()

        # if this is the 2nd time we are calling this, don't want our compiled .class to be deleted
        if not os.path.exists(tempdir):
            self.artifact.create_temp_dir(True)

        return tempdir

    def setup_main_method(self):
        if self.artifact.args.has_key('main'):
            return self.artifact.args['main']
        else:
            return os.path.splitext(os.path.basename(self.artifact.name))[0]

class JavacFilter(JavaFilter):
    ALIASES = ['javac']
    BINARY = True
    EXECUTABLE = "javac"
    INPUT_EXTENSIONS = [".java"]
    OUTPUT_EXTENSIONS = [".class"]
    VERSION_COMMAND = "java -version"

    def process(self):
        # Compile the code
        command = self.compile_command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(proc.returncode, stdout)

        # Copy compiled .class file to where it should live
        tempdir = self.artifact.temp_dir()
        compiled_file = os.path.join(tempdir, self.artifact.canonical_filename())
        shutil.copyfile(compiled_file, self.artifact.filepath())

class JavadocsJsonFilter(DexyFilter):
    ALIASES = ['javadoc', 'javadocs']

    def nested_subclasses(self, j, qualified_class_name, nest=None, indent = 0):
        if not nest:
            nest = [[indent, qualified_class_name]]
        package_name, x, class_name = qualified_class_name.rpartition(".")
        if j['packages'][package_name]['classes'][class_name].has_key('subclasses'):
            subclasses = j['packages'][package_name]['classes'][class_name]['subclasses']
            for qualified_subclass_name in subclasses:
                nest.append([indent+1, qualified_subclass_name])
                subclass_package_name, x, subclass_name = qualified_subclass_name.rpartition(".")
                if j['packages'][subclass_package_name]['classes'][subclass_name].has_key('subclasses'):
                    self.nested_subclasses(j, qualified_subclass_name, nest, indent+1)
        return nest

    def process_text(self, input_text):
        j = json.loads(input_text)

        html_formatter = HtmlFormatter()
        latex_formatter = LatexFormatter()
        lexer = JavaLexer()

        for p in j['packages']:
            for k in j['packages'][p]['classes'].keys():

                klass = j['packages'][p]['classes'][k]
                if klass.has_key('superclass'):
                    superclass_package, _, superclass_name = klass['superclass'].rpartition(".")
                    if j['packages'].has_key(superclass_package):
                        if not j['packages'][superclass_package]['classes'].has_key(superclass_name):
                            self.log.warn("Can't find %s in package %s" % (superclass_name, superclass_package))
                        else:
                            superclass = j['packages'][superclass_package]['classes'][superclass_name]
                            if not superclass.has_key('subclasses'):
                                superclass['subclasses'] = []

                            superclass['subclasses'].append("%s.%s" % (p, k))

                if klass.has_key('interfaces'):
                    for iface in klass['interfaces']:
                        iface_package, _, iface_name = iface.rpartition(".")
                        if j['packages'].has_key(iface_package):
                            if not j['packages'][iface_package]['classes'].has_key(iface_name):
                                self.log.warn("Can't find", iface_name, "in package", iface_package)
                            else:
                                iface_dict = j['packages'][iface_package]['classes'][iface_name]
                                if not iface_dict.has_key('implementers'):
                                    iface_dict['implementers'] = []
                                iface_dict['implementers'].append(klass['qualified-name'])
                if klass.has_key('source') and klass['source']:
                    source = klass['source']
                    j['packages'][p]['classes'][k]['source-html'] = str(highlight(source, lexer, html_formatter))
                    j['packages'][p]['classes'][k]['source-latex'] = str(highlight(source, lexer, latex_formatter))

                for m in j['packages'][p]['classes'][k]['methods'].keys():
                    source = j['packages'][p]['classes'][k]['methods'][m]['source']
                    if not source:
                        source = ""

                    # TODO - try running comment text through Textile
                    # interpreter for HTML and LaTeX options
                    comment = j['packages'][p]['classes'][k]['methods'][m]['comment-text']


                    j['packages'][p]['classes'][k]['methods'][m]['comment'] = comment
                    j['packages'][p]['classes'][k]['methods'][m]['source'] = source
                    j['packages'][p]['classes'][k]['methods'][m]['source-html'] = str(highlight(source, lexer, html_formatter))
                    j['packages'][p]['classes'][k]['methods'][m]['source-latex'] = str(highlight(source, lexer, latex_formatter))

                for c in j['packages'][p]['classes'][k]['constructors'].keys():
                    source = j['packages'][p]['classes'][k]['constructors'][c]['source']
                    if not source:
                        source = ""

                    # TODO - try running comment text through Textile
                    # interpreter for HTML and LaTeX options
                    comment = j['packages'][p]['classes'][k]['constructors'][c]['comment-text']

                    j['packages'][p]['classes'][k]['constructors'][c]['comment'] = comment
                    j['packages'][p]['classes'][k]['constructors'][c]['source'] = source
                    j['packages'][p]['classes'][k]['constructors'][c]['source-html'] = str(highlight(source, lexer, html_formatter))
                    j['packages'][p]['classes'][k]['constructors'][c]['source-latex'] = str(highlight(source, lexer, latex_formatter))

        return json.dumps(j)



from dexy.dexy_filter import DexyFilter
import json
import mwclient
import os

class MediaWikiFilter(DexyFilter):
    """Posts content to a MediaWiki instance"""

    MW_CONFIG_FILE = "mw-config.json"
    OUTPUT_EXTENSIONS = ['.json']
    ALIASES = ["mw"]
    MEDIA_FILE_EXTENSIONS = ['.png', '.jpg']
    INPUT_EXTENSIONS = [".txt"] + MEDIA_FILE_EXTENSIONS

    def load_mw_config(self):
        if not os.path.exists(self.MW_CONFIG_FILE):
            raise Exception("Could not find config file called %s" % self.MW_CONFIG_FILE)
        f = open(self.MW_CONFIG_FILE, "r")
        self.config = json.load(f)
        f.close()

    def process(self):
        self.load_mw_config()

        if self.config.has_key("PATH"):
            path = self.config['PATH']
        else:
            path = '/'
        site = mwclient.Site(self.config['HOST'], path)

        site.login(self.config['USERNAME'], self.config['PASSWORD'])

        if (self.artifact.input_ext in self.MEDIA_FILE_EXTENSIONS):
            # upload an image
            f = open(self.artifact.previous_artifact_filepath, 'rb')
            self.artifact.data_dict['1'] = json.dumps(site.upload(f, self.artifact.previous_canonical_filename, 'Uploaded via Dexy', ignore = True))
            f.close()
        elif (self.artifact.input_ext == '.txt'):
            page = site.Pages["/%s" % self.artifact.previous_canonical_filename]
            page.edit()
            comment = 'Uploaded via Dexy'
            result = page.save(self.artifact.input_text(), comment)
            self.artifact.data_dict['1'] = json.dumps(result)
        else:
            raise Exception("unexpected file extension %s" % self.artifact.ext)




from dexy.constants import Constants
import codecs
import dexy.database
import dexy.introspect
import json
import logging
import logging.handlers
import os

def load_batch_info(batch_id, logsdir=Constants.DEFAULT_LDIR):
    with codecs.open(batch_info_filename(batch_id, logsdir), "r", encoding="utf-8") as f:
        return json.load(f)

def save_batch_info(batch_id, batch_info, logsdir=Constants.DEFAULT_LDIR):
    with codecs.open(batch_info_filename(batch_id, logsdir), "w", encoding="utf-8") as f:
        json.dump(batch_info, f, sort_keys = True, indent = 4)

def batch_info_filename(batch_id, logsdir):
    return os.path.join(logsdir, "batch-%05d.json" % batch_id)

def get_db(db_classname=Constants.DEFAULT_DBCLASS, **kwargs):
    # TODO cache the database classes list somewhere
    database_classes = dexy.introspect.database_classes()
    db_class = database_classes[db_classname]
    return db_class(**kwargs)

def get_log(
        name=Constants.DEFAULT_LOGGER_NAME, # Name of the logger
        logsdir=Constants.DEFAULT_LDIR, # Directory to store the logfile
        logfile=Constants.DEFAULT_LFILE, # Filename of logfile
        loglevel=Constants.DEFAULT_LOGLEVEL # Log level to use
    ):
    """
    Get a log.
    """
    log = logging.getLogger(Constants.DEFAULT_LOGGER_NAME)
    log.propagate = 0

    if len(log.handlers) == 0:
        # Set up handlers and formatters.
        try:
            log.setLevel(Constants.LOGLEVELS[loglevel])
        except KeyError:
            msg = "You requested log level '%s', valid values are %s" % (loglevel, ",".join(Constants.LOGLEVELS.keys()))
            raise Exception(msg)

        if logsdir and logfile:
            logfile = os.path.join(logsdir, logfile)
            handler = logging.handlers.RotatingFileHandler(logfile)
        else:
            handler = logging.StreamHandler() # log to sys.stderr

        log.addHandler(handler)
        formatter = logging.Formatter(Constants.DEFAULT_LOGFORMAT)
        handler.setFormatter(formatter)

    if name == Constants.DEFAULT_LOGGER_NAME:
        return log
    else:
        log2 = logging.getLogger(name)
        log2.handlers = log.handlers
        if name == "dexy.controller":
            log2.propagate = 0
        return log2

def remove_all_handlers(log):
    for h in log.handlers:
        log.removeHandler(h)

def print_string_diff(str1, str2):
    msg = ""
    for i, c1 in enumerate(str1):
        if len(str2) > i:
            c2 = str2[i]
            if c1 == c2:
                flag = ""
            else:
                flag = " <---"
            if ord(c1) > ord('a') and ord(c2) > ord('a'):
                msg = msg + "\n%5d: %s\t%s\t\t%s\t%s %s" % (i, c1, c2,
                                              ord(c1), ord(c2), flag)
            else:
                msg = msg + "\n%5d:  \t \t\t%s\t%s %s" % (i, ord(c1),
                                              ord(c2), flag)
        else:
            flag = "<---"
            msg = msg + "\n%5d:  \t \t\t%s %s" % (i, ord(c1), flag)
    return msg

# Based on http://nedbatchelder.com/code/utilities/wh.py
def command_exists(cmd_name):
    path = os.environ["PATH"]
    if ";" in path:
        path = filter(None, path.split(";"))
    else:
        path = filter(None, path.split(":"))

    if os.environ.has_key("PATHEXT"):
        # Present on windows, returns e.g. '.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC'
        pathext = os.environ["PATHEXT"]
        pathext = filter(None, pathext.split(";"))
    else:
        # Not windows, look for exact command name.
        pathext = [""]

    cmd_found = False
    for d in path:
        for e in pathext:
            filepath = os.path.join(d, cmd_name + e)
            if os.path.exists(filepath):
                cmd_found = True
                break
    return cmd_found

#http://code.activestate.com/recipes/148061-one-liner-word-wrap-function/
def wrap_text(text, width):
    """
    A word-wrap function that preserves existing line breaks
    and most spaces in the text. Expects that existing line
    breaks are posix newlines (\n).
    """
    return reduce(lambda line, word, width=width: '%s%s%s' %
             (line,
               ' \n'[(len(line)-line.rfind('\n')-1
                     + len(word.split('\n',1)[0]
                          ) >= width)],
               word),
              text.split(' ')
             )



from dexy.dexy_filter import DexyFilter
from lxml import etree
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.formatters.latex import LatexFormatter
from pygments.lexers.web import XmlLexer
import json


class XmlSectionFilter(DexyFilter):
    """
    A filter for breaking an XML file up into various sections, referenced by
    xpaths, and optionally syntax highlighting the various sections to allow
    easy access to any subsection of an XML document.
    """
    INPUT_EXTENSIONS = [".xml"]
    OUTPUT_EXTENSIONS = [".json"]
    ALIASES = ["xxml"]
    FINAL = False

    def process_text(self, input_text):
        lexer = XmlLexer()
        html_formatter = HtmlFormatter(lineanchors=self.artifact.web_safe_document_key())
        latex_formatter = LatexFormatter()

        output = {}

        root = etree.fromstring(input_text)
        tree = root.getroottree()

        for element in tree.iter("*"):
            xpath = tree.getpath(element)
            source = etree.tostring(element, pretty_print=True).strip()

            # Convert _Attrib to dict
            attributes = dict((k, v) for k, v in element.attrib.iteritems())

            element_info = {
                    "attributes" : attributes,
                    "lineno" : element.sourceline,
                    "prefix" : element.prefix,
                    "source" : source,
                    "source-html" : highlight(source, lexer, html_formatter),
                    "source-latex" : highlight(source, lexer, latex_formatter),
                    "tag" : element.tag,
                    "tail" : element.tail,
                    "text" : element.text
                }
            output[xpath] = element_info

            if attributes.has_key("name"):
                key = "%s:%s" % (element.tag, attributes["name"])
                output[key] = element_info

        return json.dumps(output, indent=4, sort_keys=True)



import dexy.utils
import platform
import subprocess

class DexyFilterException(Exception):
    pass

class DexyFilter(object):
    """
    This is the main DexyFilter class. To make custom filters you should
    subclass this and override the process() method. You may also want to
    specify INPUT_EXTENSIONS and OUTPUT_EXTENSIONS. You must define unique
    ALIASES in each handler, use java-style namespacing, e.g. com.abc.alias
    """
    ALIASES = ['dexy']
    BINARY = False
    FINAL = None
    IGNORE_ERRORS = False # Set to True for filters where proc.returncode doesn't behave well
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".*"]
    TAGS = [] # Descriptive keywords about the filter.
    VERSION_COMMAND = None
    WINDOWS_VERSION_COMMAND = None

    @classmethod
    def filters(self):
        """Lists available filter classes."""
        pass

    @classmethod
    def executable(self):
        """A standard way of specifying a command line executable. For usage
        example see stdout filter. This does not need to be used, and is not
        relevant for many filters, but is intended to allow introspection for
        those which do use it."""
        if platform.system() == 'Windows' and hasattr(self, 'WINDOWS_EXECUTABLE'):
            return self.WINDOWS_EXECUTABLE
        else:
            if hasattr(self, 'EXECUTABLE'):
                return self.EXECUTABLE
            elif hasattr(self, 'EXECUTABLES'):
                # Allows you to specify multiple options for an executable and,
                # at runtime, use whichever one is present on the system. The
                # first listed executable to be found is the one used.
                return self.find_present_executable()


    @classmethod
    def find_present_executable(klass):
        # determine which executable to use
        for exe in klass.EXECUTABLES:
            if klass.executable_present(exe):
                return exe
                break
        return None

    @classmethod
    def executable_present(klass, exe=None):
        """Determine whether the specified executable is available."""
        if not exe:
            exe = klass.executable()

        if exe:
            cmd = exe.split()[0] # remove any --arguments
            return dexy.utils.command_exists(cmd)
        else:
            # why true? because there's nothing to run?
            return True

    @classmethod
    def version_command(klass):
        if platform.system() == 'Windows':
            return klass.WINDOWS_VERSION_COMMAND or klass.VERSION_COMMAND
        else:
            return klass.VERSION_COMMAND

    @classmethod
    def version(klass, log=None):
        vc = klass.version_command()

        if vc:
            # TODO make custom env available here...
            proc = subprocess.Popen(vc, shell=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)
            stdout, stderr = proc.communicate()

            if proc.returncode > 0:
                err_msg = """An error occurred running %s""" % vc
                if log:
                    log.debug(err_msg)
                return False
            else:
                return stdout.strip().split("\n")[0]
        else:
            return None

    @classmethod
    def output_file_extension(klass, ext, key, next_input_extensions=None):
        out_ext = None

        if set([ext, ".*"]).isdisjoint(set(klass.INPUT_EXTENSIONS)):
            exception_text = """Error in %s for %s. Extension %s is not supported.
            Supported extensions are: %s""" % (klass.__name__, key, ext, ', '.join(klass.INPUT_EXTENSIONS))
            raise Exception(exception_text)

        if ".*" in klass.OUTPUT_EXTENSIONS:
            out_ext = ext
        else:
            if next_input_extensions and not ".*" in next_input_extensions:
                for e in klass.OUTPUT_EXTENSIONS:
                    if e in next_input_extensions:
                        out_ext = e

                if not out_ext:
                  err_str = "unable to find one of %s in %s for %s"
                  prev_out = ", ".join(klass.OUTPUT_EXTENSIONS)
                  next_in = ", ".join(next_input_extensions)
                  err_str = err_str % (prev_out, next_in, key)
                  raise Exception(err_str)
            else:
                out_ext = klass.OUTPUT_EXTENSIONS[0]
        return out_ext

    @classmethod
    def enabled(self):
        """Allow filters to be disabled."""
        return True

    def process(self):
        """This is the method that does the "work" of the handler, that is
        filtering the input and producing output. This method can be overridden
        in a subclass, or one of the convenience methods named below can be
        implemented and will be delegated to. If more than 1 convenience method
        is implemented then an exception will be raised."""
        method_used = None

        if hasattr(self, "process_text"):
            if method_used:
                raise Exception("%s has already been called" % method_used)
            if len(self.artifact.input_data_dict.keys()) > 1:
                raise Exception("""You have passed input with multiple sections
                                to the %s handler. This handler does not preserve
                                sections. Either remove sectioning or add a call
                                to the join filter before this handler.""" % self.__class__.__name__)
            input_text = self.artifact.input_text()
            output_text = self.process_text(input_text)
            self.artifact.data_dict['1'] = output_text
            method_used = "process_text"

        if hasattr(self, "process_dict"):
            if method_used:
                raise Exception("%s has already been called" % method_used)
            input_dict = self.artifact.input_data_dict
            output_dict = self.process_dict(input_dict)
            self.artifact.data_dict = output_dict
            method_used = "process_dict"

        if hasattr(self, "process_text_to_dict"):
            if method_used:
                raise Exception("%s has already been called" % method_used)
            input_text = self.artifact.input_text()
            output_dict = self.process_text_to_dict(input_text)
            self.artifact.data_dict = output_dict
            method_used = "process_text_to_dict"

        if not method_used:
            # This code implements the neutral 'dexy' handler.
            self.artifact.data_dict = self.artifact.input_data_dict
            method_used = "process"

        return method_used
 



from dexy.artifact import Artifact
from ordereddict import OrderedDict
import codecs
import json
import os

class FileSystemJsonArtifact(Artifact):
    """Artifact which persists data by writing to the file system and using
    JSON for serializing metadata (default type of Artifact)"""

    # Metadata
    def meta_filename(self):
        return "%s-meta.json" % (self.hashstring)

    def meta_filepath(self):
        return os.path.join(self.artifacts_dir, self.meta_filename())

    def write_dict_to_file(self, data_dict, filepath):
        with open(filepath, "wb") as f:
            json.dump(data_dict, f)

    def read_dict_from_file(self, filepath):
        with open(filepath, "rb") as f:
            return json.load(f)

    def save_meta(self):
        m = {}
        attrs_to_persist = set(self.META_ATTRS + self.HASH_WHITELIST) - set(['input_data_dict', 'inputs'])
        for a in attrs_to_persist:
            if hasattr(self, a):
                v = getattr(self, a)
                m[a] = v

        m['inputs'] = self.input_hashes()

        self.write_dict_to_file(m, self.meta_filepath())

    def load_meta(self):
        m = self.read_dict_from_file(self.meta_filepath())

        self._inputs = dict((k, self.__class__.retrieve(h)) for (k, h) in m.pop('inputs').iteritems())

        for k, v in m.iteritems():
            setattr(self, k, v)

        # We only store filter name, not filter class, need to retrieve class from name
        if hasattr(self, "filter_name") and not hasattr(self, "filter_class"):
            self.filter_class = [k for n,k in self.FILTERS.iteritems() if k.__name__ == self.filter_name][0]

    # Input
    def load_input(self):
        """Load input data into memory, if applicable."""
        if self.is_loaded():
            return

        if self.binary_input:
            #not loading non-binary input
            pass
        elif self.initial:
            #initial artifact has no input
            pass
        elif self.additional:
            #additional artifact has no input
            pass
        elif len(self.input_data_dict) > 0:
            #we already have input data in memory
            pass
        elif not hasattr(self, 'previous_cached_output_filepath'):
            #no previous cached output, can't load
            pass
        else:
            with open(self.previous_cached_output_filepath, "rb") as f:
                self.input_data_dict = self.convert_numbered_dict_to_ordered_dict(json.load(f))

    # Output
    def cached_output_filename(self):
        return "%s-output.json" % (self.hashstring)

    def cached_output_filepath(self):
        return os.path.join(self.artifacts_dir, self.cached_output_filename())

    def is_output_cached(self):
        # TODO add checksums to verify data hasn't changed
        if self.binary_output:
            return self.is_canonical_output_cached()
        else:
            return self.is_json_output_cached() and self.is_canonical_output_cached()

    def is_json_output_cached(self):
        fp = self.cached_output_filepath()
        return os.path.isfile(fp) and (os.path.getsize(fp) > 0)

    def is_canonical_output_cached(self):
        fp = self.filepath()
        return os.path.isfile(fp) and (os.path.getsize(fp) > 0)

    def save_output(self):
        if not self.is_complete():
            raise Exception("should not be calling save_output unless artifact is complete")

        if not self.binary_output:
            if not self.data_dict or len(self.data_dict) == 0:
                # Our filter has written directly to an output file
                # We need to load this into memory first
                self.data_dict = OrderedDict()
                with codecs.open(self.filepath(), 'r', encoding="utf-8") as f:
                    data = f.read()
                self.data_dict['1'] = data
            else:
                # Write the canonical output file.
                with codecs.open(self.filepath(), 'w', encoding="utf-8") as f:
                    f.write(self.output_text())

            # Write the JSON output file.
            output_dict = self.convert_data_dict_to_numbered_dict()
            self.write_dict_to_file(output_dict, self.cached_output_filepath())

    def load_output(self):
        if not self.is_complete():
            raise Exception("should not be calling load_output unless artifact is complete")

        if not self.binary_output:
            output_dict = self.read_dict_from_file(self.cached_output_filepath())
            self.data_dict = self.convert_numbered_dict_to_ordered_dict(output_dict)




#-----------------------------------------------
#aplotter.py - ascii art function plotter
#Copyright (c) 2006, Imri Goldberg
#All rights reserved.
# http://www.algorithm.co.il/blogs/ascii-plotter/
#
#Redistribution and use in source and binary forms,
#with or without modification, are permitted provided
#that the following conditions are met:
#
#    * Redistributions of source code must retain the
#        above copyright notice, this list of conditions
#        and the following disclaimer.
#    * Redistributions in binary form must reproduce the
#        above copyright notice, this list of conditions
#        and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.
#    * Neither the name of the <ORGANIZATION> nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
#LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#-----------------------------------------------

import math


EPSILON = 0.000001

def transposed(mat):
    result = []
    for i in xrange(len(mat[0])):
        result.append([x[i] for x in mat])
    return result

def y_reversed(mat):    
    result = []
    for i in range(len(mat)):
        result.append(list(reversed(mat[i])))
    return result

def sign(x):
    if 0<x:
        return 1
    if 0 == x:
        return 0
    return -1

class Plotter(object):

    class PlotData(object):
        def __init__(self, x_size, y_size, min_x, max_x, min_y, max_y, x_mod, y_mod):
            self.x_size = x_size
            self.y_size = y_size
            self.min_x = min_x
            self.max_x = max_x
            self.min_y = min_y
            self.max_y = max_y
            self.x_mod = x_mod
            self.y_mod = y_mod

            self.x_step = float(max_x - min_x)/float(self.x_size)
            self.y_step = float(max_y - min_y)/float(self.y_size)
            self.inv_x_step = 1/self.x_step
            self.inv_y_step = 1/self.y_step

            self.ratio = self.y_step / self.x_step
        def __repr__(self):
            s = "size: %s, bl: %s, tr: %s, step: %s" % ((self.x_size, self.y_size), (self.min_x, self.min_y), (self.max_x, self.max_y),
                                                         (self.x_step, self.y_step))
            return s
    
    def __init__(self, **kwargs):

        self.x_size = kwargs.get("x_size", 80)
        self.y_size = kwargs.get("y_size", 20)

        self.will_draw_axes = kwargs.get("draw_axes", True)

        self.new_line = kwargs.get("newline", "\n")

        self.dot = kwargs.get("dot", "*")

        self.plot_slope = kwargs.get("plot_slope", True)

        self.x_margin = kwargs.get("x_margin", 0.05)
        self.y_margin = kwargs.get("y_margin", 0.1)

        self.will_plot_labels = kwargs.get("plot_labels", True)

    @staticmethod
    def get_symbol_by_slope(slope, default_symbol):
        draw_symbol = default_symbol
        if slope > math.tan(3*math.pi/8):
            draw_symbol = "|"
        elif slope > math.tan(math.pi/8) and slope < math.tan(3*math.pi/8):
            draw_symbol = "/"
        elif abs(slope) < math.tan(math.pi/8):
            draw_symbol = "-"
        elif slope < math.tan(-math.pi/8) and slope > math.tan(-3*math.pi/8):
            draw_symbol = "\\"
        elif slope < math.tan(-3*math.pi/8):
            draw_symbol = "|"
        return draw_symbol    


    def plot_labels(self, output_buffer, plot_data):
        if plot_data.y_size < 2:
            return

        margin_factor = 1

        do_plot_x_label = True
        do_plot_y_label = True

        x_str = "%+g"
        if plot_data.x_size < 16:
            do_plot_x_label = False
        elif plot_data.x_size < 23:
            x_str = "%+.2g" 

        y_str = "%+g"    
        if plot_data.x_size < 8:
            do_plot_y_label = False
        elif plot_data.x_size < 11:
            y_str = "%+.2g"
            
        act_min_x = (plot_data.min_x + plot_data.x_mod*margin_factor)
        act_max_x = (plot_data.max_x - plot_data.x_mod*margin_factor)
        act_min_y = (plot_data.min_y + plot_data.y_mod*margin_factor)
        act_max_y = (plot_data.max_y - plot_data.y_mod*margin_factor)

        if abs(act_min_x) < 1:
            min_x_str = "%+.2g" % act_min_x
        else:
            min_x_str = x_str % act_min_x

        if abs(act_max_x) < 1:
            max_x_str = "%+.2g" % act_max_x
        else:
            max_x_str = x_str % act_max_x
        
        if abs(act_min_y) < 1:
            min_y_str = "%+.2g" % act_min_y
        else:
            min_y_str = y_str % act_min_y

        if abs(act_max_y) < 1:
            max_y_str = "%+.2g" % act_max_y
        else:
            max_y_str = y_str % act_max_y
                         
        min_x_coord = self.get_coord(act_min_x,plot_data.min_x,plot_data.x_step)
        max_x_coord = self.get_coord(act_max_x,plot_data.min_x,plot_data.x_step)
        min_y_coord = self.get_coord(act_min_y,plot_data.min_y,plot_data.y_step)
        max_y_coord = self.get_coord(act_max_y,plot_data.min_y,plot_data.y_step)
                                 

        #print plot_data
        
        y_zero_coord = self.get_coord(0, plot_data.min_y, plot_data.y_step)

        if plot_data.min_x < 0 and plot_data.max_x > 0:
             x_zero_coord = self.get_coord(0, plot_data.min_x, plot_data.x_step)
        else:
             pass

        try:
            output_buffer[x_zero_coord][min_y_coord] = "+"
            output_buffer[x_zero_coord][max_y_coord] = "+"
            output_buffer[min_x_coord][y_zero_coord] = "+"
            output_buffer[max_x_coord][y_zero_coord] = "+"
        except:
             pass

        if do_plot_x_label:

            for i,c in enumerate(min_x_str):
                output_buffer[min_x_coord+i][y_zero_coord-1] = c
            for i,c in enumerate(max_x_str):
                output_buffer[max_x_coord+i-len(max_x_str)][y_zero_coord-1] = c

        if do_plot_y_label:

            for i,c in enumerate(max_y_str):
                output_buffer[x_zero_coord+i][max_y_coord] = c
            for i,c in enumerate(min_y_str):
                output_buffer[x_zero_coord+i][min_y_coord] = c
            

        
        
    
    def plot_data(self, xy_seq, output_buffer, plot_data):
        if self.plot_slope:
            xy_seq = list(xy_seq)
            #sort according to the x coord
            xy_seq.sort(key = lambda c: c[0])
            prev_p = xy_seq[0]
            e_xy_seq = enumerate(xy_seq)
            e_xy_seq.next()
            for i,(x,y) in e_xy_seq:
                draw_symbol = self.dot
                line_drawn = self.plot_line(prev_p, (x,y), output_buffer, plot_data)
                prev_p = (x,y)
                if not line_drawn:
                    if i > 0 and i < len(xy_seq)-1:
                        px,py = xy_seq[i-1]
                        nx,ny = xy_seq[i+1]

                        if abs(nx-px) > EPSILON:
                            slope = (1.0/plot_data.ratio)*(ny-py)/(nx-px)
                            draw_symbol = self.get_symbol_by_slope(slope, draw_symbol)
                    if x < plot_data.min_x or x >= plot_data.max_x or y < plot_data.min_y or y >= plot_data.max_y:
                        continue
                    
                    x_coord = self.get_coord(x, plot_data.min_x, plot_data.x_step)
                    y_coord = self.get_coord(y, plot_data.min_y, plot_data.y_step)            
                    if x_coord >= 0 and x_coord < len(output_buffer) and y_coord >= 0 and y_coord < len(output_buffer[0]):
                        if self.draw_axes:
                            if y_coord == self.get_coord(0, plot_data.min_y, plot_data.y_step) and draw_symbol == "-":
                                draw_symbol = "="
                        output_buffer[x_coord][y_coord] = draw_symbol
        else:
            for x,y in xy_seq:
                if x < plot_data.min_x or x >= plot_data.max_x or y < plot_data.min_y or y >= plot_data.max_y:
                    continue
                x_coord = self.get_coord(x, plot_data.min_x, plot_data.x_step)
                y_coord = self.get_coord(y, plot_data.min_y, plot_data.y_step)
                if x_coord >= 0 and x_coord < len(output_buffer) and y_coord > 0 and y_coord < len(output_buffer[0]):
                    output_buffer[x_coord][y_coord] = self.dot


    def plot_line(self, start, end, output_buffer, plot_data):

        start_coord = self.get_coord(start[0], plot_data.min_x, plot_data.x_step), self.get_coord(start[1], plot_data.min_y, plot_data.y_step)
        end_coord = self.get_coord(end[0], plot_data.min_x, plot_data.x_step), self.get_coord(end[1], plot_data.min_y, plot_data.y_step)

        x0,y0 = start_coord
        x1,y1 = end_coord
        if (x0,y0) == (x1,y1):
            return True    
        
        clipped_line = clip_line(start, end, (plot_data.min_x, plot_data.min_y), (plot_data.max_x, plot_data.max_y))
        if clipped_line != None:
            start,end = clipped_line
        else:
            return False
        start_coord = self.get_coord(start[0], plot_data.min_x, plot_data.x_step), self.get_coord(start[1], plot_data.min_y, plot_data.y_step)
        end_coord = self.get_coord(end[0], plot_data.min_x, plot_data.x_step), self.get_coord(end[1], plot_data.min_y, plot_data.y_step)

        x0,y0 = start_coord
        x1,y1 = end_coord
        if (x0,y0) == (x1,y1):
            return True
        x_zero_coord = self.get_coord(0, plot_data.min_x, plot_data.x_step)
        y_zero_coord = self.get_coord(0, plot_data.min_y, plot_data.y_step)   

        if start[0]-end[0] == 0:
            draw_symbol = "|"
        else:
            slope = (1.0/plot_data.ratio)*(end[1]-start[1])/(end[0]-start[0])
            draw_symbol = self.get_symbol_by_slope(slope, self.dot)
        try:

            delta = x1-x0, y1-y0
            if abs(delta[0])>abs(delta[1]):
                s = sign(delta[0])
                slope = float(delta[1])/delta[0]
                for i in range(0,abs(int(delta[0]))):
                    cur_draw_symbol = draw_symbol
                    x = i*s
                    cur_y = int(y0+slope*x)
                    if self.draw_axes and cur_y == y_zero_coord and draw_symbol == "-":
                        cur_draw_symbol = "="
                    output_buffer[x0+x][cur_y] = cur_draw_symbol
                
                
            else:
                s = sign(delta[1])
                slope = float(delta[0])/delta[1]
                for i in range(0,abs(int(delta[1]))):
                    y = i*s
                    cur_draw_symbol = draw_symbol
                    cur_y = y0+y
                    if self.draw_axes and cur_y == y_zero_coord and draw_symbol == "-":
                        cur_draw_symbol = "="
                    output_buffer[int(x0+slope*y)][cur_y] = cur_draw_symbol
        except:
            print start, end
            print start_coord, end_coord
            print plot_data
            raise

        return False            
        
        
    def plot_single(self, seq, min_x = None, max_x = None, min_y = None, max_y = None):
        return self.plot_double(range(len(seq)),seq, min_x, max_x, min_y, max_y)
        



    def plot_double(self, x_seq, y_seq, min_x = None, max_x = None, min_y = None, max_y = None):
        if min_x == None:
            min_x = min(x_seq)
        if max_x == None:
            max_x = max(x_seq)
        if min_y == None:
            min_y = min(y_seq)
        if max_y == None:
            max_y = max(y_seq)

        if max_y == min_y:
            max_y += 1

        x_mod = (max_x-min_x)*self.x_margin
        y_mod = (max_y-min_y)*self.y_margin
        min_x-=x_mod
        max_x+=x_mod
        min_y-=y_mod
        max_y+=y_mod


        plot_data = self.PlotData(self.x_size, self.y_size, min_x, max_x, min_y, max_y, x_mod, y_mod)

        output_buffer = [[" "]*self.y_size for i in range(self.x_size)]

        if self.will_draw_axes:
            self.draw_axes(output_buffer, plot_data)

        self.plot_data(zip(x_seq, y_seq), output_buffer, plot_data)

        if self.will_plot_labels:
            self.plot_labels(output_buffer, plot_data)

        trans_result = transposed(y_reversed(output_buffer))

        result = self.new_line.join(["".join(row) for row in trans_result])
        return result

    def draw_axes(self, output_buffer, plot_data):
        
        
        draw_x = False
        draw_y = False

        if plot_data.min_x <= 0 and plot_data.max_x > 0:
            draw_y = True
            zero_x = self.get_coord(0, plot_data.min_x, plot_data.x_step)
            for y in xrange(plot_data.y_size):
                output_buffer[zero_x][y] = "|"
                
        if plot_data.min_y <= 0 and plot_data.max_y > 0:
            draw_x = True
            zero_y = self.get_coord(0, plot_data.min_y, plot_data.y_step)    
            for x in xrange(plot_data.x_size):
                output_buffer[x][zero_y] = "-"

        if draw_x and draw_y:
            output_buffer[zero_x][zero_y] = "+"
        
        
    @staticmethod
    def get_coord(val, min, step):
        result = int((val - min)/step)
        return result

def clip_line(line_pt_1, line_pt_2, rect_bottom_left, rect_top_right):
    ts = [0.0,1.0]
    if line_pt_1[0] == line_pt_2[0]:
        return ((line_pt_1[0], max(min(line_pt_1[1], line_pt_2[1]), rect_bottom_left[1])),
                (line_pt_1[0], min(max(line_pt_1[1], line_pt_2[1]), rect_top_right[1])))
    if line_pt_1[1] == line_pt_2[1]:
        return ((max(min(line_pt_1[0], line_pt_2[0]), rect_bottom_left[0]), line_pt_1[1]),
                (min(max(line_pt_1[0], line_pt_2[0]), rect_top_right[0]), line_pt_1[1]))

    if ((rect_bottom_left[0] <= line_pt_1[0] and line_pt_1[0] < rect_top_right[0]) and
        (rect_bottom_left[1] <= line_pt_1[1] and line_pt_1[1] < rect_top_right[1]) and
        (rect_bottom_left[0] <= line_pt_2[0] and line_pt_2[0] < rect_top_right[0]) and
        (rect_bottom_left[1] <= line_pt_2[1] and line_pt_2[1] < rect_top_right[1])):
        return line_pt_1, line_pt_2

    ts.append( float(rect_bottom_left[0]-line_pt_1[0])/(line_pt_2[0]-line_pt_1[0]) )
    ts.append( float(rect_top_right[0]-line_pt_1[0])/(line_pt_2[0]-line_pt_1[0]) )
    ts.append( float(rect_bottom_left[1]-line_pt_1[1])/(line_pt_2[1]-line_pt_1[1]) )
    ts.append( float(rect_top_right[1]-line_pt_1[1])/(line_pt_2[1]-line_pt_1[1]) )
    
    ts.sort()
    if ts[2] < 0 or ts[2] >= 1 or ts[3] < 0 or ts[2]>= 1:
        return None
    result = [(pt_1 + t*(pt_2-pt_1)) for t in (ts[2],ts[3]) for (pt_1, pt_2) in zip(line_pt_1, line_pt_2)]
    return (result[0],result[1]), (result[2], result[3])
    


def plot(*args,**flags):
    limit_flags_names = set(["min_x","min_y","max_x","max_y"])
    limit_flags = dict([(n,flags[n]) for n in limit_flags_names & set(flags)])
    settting_flags = dict([(n,flags[n]) for n in set(flags) - limit_flags_names])
    
    if len(args) == 1:
        p = Plotter(**settting_flags)
        return p.plot_single(args[0],**limit_flags)
    elif len(args) == 2:
        p = Plotter(**settting_flags)
        return p.plot_double(args[0],args[1],**limit_flags)
    else:
        raise NotImplementedError("can't draw multiple graphs yet")
    
__all__ = ["Plotter","plot"]

